####################################################################
#                                                                  #
#    ensemble_md,                                                  #
#    a python package for running GROMACS simulation ensembles     #
#                                                                  #
#    Written by Wei-Tse Hsu <wehs7661@colorado.edu>                #
#    Copyright (c) 2022 University of Colorado Boulder             #
#                                                                  #
####################################################################

class ParameterError(Exception):
    """Error raised when detecting improperly specified parameters in the YAML file."""


class ParseError(Exception):
    """Error raised when parsing of a file failed. Modified from GromacsWrapper."""



"""A repository for developing ensemble simulation methods"""

# Add imports here
# from .ensemble_md import canvas  # noqa: ABS101

# Handle versioneer
from ._version import get_versions  # noqa: ABS101

versions = get_versions()
__version__ = versions["version"]
__git_revision__ = versions["full-revisionid"]
del get_versions, versions



####################################################################
#                                                                  #
#    ensemble_md,                                                  #
#    a python package for running GROMACS simulation ensembles     #
#                                                                  #
#    Written by Wei-Tse Hsu <wehs7661@colorado.edu>                #
#    Copyright (c) 2022 University of Colorado Boulder             #
#                                                                  #
####################################################################
"""
The :code:`gmx_parser` module provides functions for parsing GROMACS files.
"""
import os
import re
import six
import logging
import warnings
from collections import OrderedDict as odict

from ensemble_md.utils import utils
from ensemble_md.utils.exceptions import ParseError


def parse_log(log_file):
    """
    This function parses a log file generated by expanded ensemble and provides
    important information especially for running new iterations in EEXE.
    Typically, there are three types of log files from an expanded ensemble simulation:

    - Case 1: The weights are equilibrating in the simulation and the equilibration was not done.
      - :code:`equil_time` should always be -1.
    - Case 2: The weights were equilibrated during the simulation.
      - :code:`equil_time` should be the time (ps) that the weights get equilibrated.
      - The final weights will just be the equilibrated weights.
    - Case 3: The weights were fixed in the simulation.
      - :code:`equil_time` should always be 0.
      - We can still find the final weights (not changed though) and the final counts.

    Parameters
    ----------
    log_file : str
        The log file to be parsed.

    Returns
    -------
    wl_delta : float
        The final Wang-Landau incementor
    weights : list
        The final list of lambda weights.
    counts : list
        The final histogram counts.
    equil_time : int or float
        The time in ps that the weights get equilibrated.
        -1 means that the weights have not been equilibrated (Case 1).
        0 means that the weights were fixed during the simulation (Case 3).
    """
    f = open(log_file, "r")
    lines = f.readlines()
    f.close()

    wl_delta_found, weights_found = False, False
    weights, counts = [], []
    case_1_3, equil_time_found = False, False

    for l in lines:  # noqa: E741
        if "n-lambdas" in l:
            N_states = int(l.split("=")[1])
        if "tinit" in l:
            tinit = float(l.split("=")[1])
        if "lmc-stats" in l:
            if l.split("=")[1].split()[0] in ["no", "No"]:
                # Case 3: The weights are fixed. -> Typically weights have been equilibrated in the previous iteration.
                fixed_bool = True
                equil_time = 0
                wl_delta = None
                wl_delta_found = True
                case_1_3 = True
            else:
                # Either Case 1 or Case 2
                fixed_bool = False
                equil_time = -1
        if "dt  " in l:
            dt = float(l.split("=")[1])

    lines.reverse()  # We find the information from the end
    n = -1
    for l in lines:  # noqa: E741
        n += 1
        if wl_delta_found is False:  # Case 1 or Case 2
            if "Wang-Landau incrementor is" in l:
                wl_delta_found = True
                wl_delta = float(l.split(":")[1])

        if weights_found is False:
            if "Count   G(in kT)" in l:
                weights_found = True  # The first occurrence would be the final weights. (We've reversed the lines!)
                for i in range(1, N_states + 1):
                    if "<<" in lines[n - i]:
                        weights.append(float(lines[n - i].split()[-3]))
                        counts.append(int(lines[n - i].split()[-4]))
                    else:
                        weights.append(float(lines[n - i].split()[-2]))
                        counts.append(int(lines[n - i].split()[-3]))
                if "Wang-Landau incrementor is" in lines[n + 1]:  # Caes 1
                    equil_time = -1
                    wl_delta_found = True
                    wl_delta = float(lines[n + 1].split(":")[1])
                    case_1_3 = True
                else:
                    # This would include Case 2 and 3
                    if fixed_bool is True:
                        case_1_3 = True
                    else:
                        case_1_3 = False

        if case_1_3 is False:  # for finding when the weights were equilibrated in Case 2
            if "Weights have equilibrated" in l:
                equil_time_found = True
                equil_step = int(l.split(":")[0].split("Step")[1])
                equil_time = equil_step * dt + tinit  # ps

        # For Case 1 and Case 3 (case_1_3 = True), we can break the loop after getting weights, counts, and wl_delta.
        if case_1_3 is True and weights_found is True and wl_delta_found is True:
            break

        # For Case 2, we can break the loop only if equil_time is found as well.
        if (
            case_1_3 is False
            and weights_found is True
            and wl_delta_found is True
            and equil_time_found is True
        ):
            break

    return wl_delta, weights, counts, equil_time


class FileUtils(object):
    """Mixin class to provide additional file-related capabilities.
    Modified from `utilities.py in GromacsWrapper <https://github.com/Becksteinlab/GromacsWrapper>`_.
    Copyright (c) 2009 Oliver Beckstein <orbeckst@gmail.com>
    """

    #: Default extension for files read/written by this class.
    default_extension = None

    def _init_filename(self, filename=None, ext=None):
        """Initialize the current filename :attr:`FileUtils.real_filename` of the object.

        Bit of a hack.

        - The first invocation must have ``filename != None``; this will set a
          default filename with suffix :attr:`FileUtils.default_extension`
          unless another one was supplied.

        - Subsequent invocations either change the filename accordingly or
          ensure that the default filename is set with the proper suffix.

        """

        extension = ext or self.default_extension
        filename = self.filename(
            filename, ext=extension, use_my_ext=True, set_default=True
        )
        #: Current full path of the object for reading and writing I/O.
        self.real_filename = os.path.realpath(filename)

    def filename(self, filename=None, ext=None, set_default=False, use_my_ext=False):
        """Supply a file name for the class object.

        Typical uses::

           fn = filename()             ---> <default_filename>
           fn = filename('name.ext')   ---> 'name'
           fn = filename(ext='pickle') ---> <default_filename>'.pickle'
           fn = filename('name.inp','pdf') --> 'name.pdf'
           fn = filename('foo.pdf',ext='png',use_my_ext=True) --> 'foo.pdf'

        The returned filename is stripped of the extension
        (``use_my_ext=False``) and if provided, another extension is
        appended. Chooses a default if no filename is given.

        Raises a ``ValueError`` exception if no default file name is known.

        If ``set_default=True`` then the default filename is also set.

        ``use_my_ext=True`` lets the suffix of a provided filename take
        priority over a default ``ext`` tension.
        """
        if filename is None:
            if not hasattr(self, "_filename"):
                self._filename = None  # add attribute to class
            if self._filename:
                filename = self._filename
            else:
                raise ValueError(
                    "A file name is required because no default file name was defined."
                )
            my_ext = None
        else:
            filename, my_ext = os.path.splitext(filename)
            if set_default:  # replaces existing default file name
                self._filename = filename
        if my_ext and use_my_ext:
            ext = my_ext
        if ext is not None:
            if ext.startswith(os.extsep):
                ext = ext[1:]  # strip a dot to avoid annoying mistakes
            if ext != "":
                filename = filename + os.extsep + ext
        return filename


class MDP(odict, FileUtils):
    """Class that represents a Gromacs mdp run input file.
    Modified from `GromacsWrapper <https://github.com/Becksteinlab/GromacsWrapper>`_.
    Copyright (c) 2009-2011 Oliver Beckstein <orbeckst@gmail.com>
    The MDP instance is an ordered dictionary.

      - *Parameter names* are keys in the dictionary.
      - *Comments* are sequentially numbered with keys Comment0001,
        Comment0002, ...
      - *Empty lines* are similarly preserved as Blank0001, ....

    When writing, the dictionary is dumped in the recorded order to a
    file. Inserting keys at a specific position is not possible.

    Currently, comments after a parameter on the same line are
    discarded. Leading and trailing spaces are always stripped.
    """

    default_extension = "mdp"
    logger = logging.getLogger("gromacs.formats.MDP")

    COMMENT = re.compile("""\s*;\s*(?P<value>.*)""")  # eat initial ws  # noqa: W605
    # see regex in cbook.edit_mdp()
    PARAMETER = re.compile(
        """
                            \s*(?P<parameter>[^=]+?)\s*=\s*  # parameter (ws-stripped), before '='  # noqa: W605
                            (?P<value>[^;]*)                # value (stop before comment=;)  # noqa: W605
                            (?P<comment>\s*;.*)?            # optional comment  # noqa: W605
                            """,
        re.VERBOSE,
    )

    def __init__(self, filename=None, autoconvert=True, **kwargs):
        """Initialize mdp structure.

        :Arguments:
          *filename*
              read from mdp file
          *autoconvert* : boolean
              ``True`` converts numerical values to python numerical types;
              ``False`` keeps everything as strings [``True``]
          *kwargs*
              Populate the MDP with key=value pairs. (NO SANITY CHECKS; and also
              does not work for keys that are not legal python variable names such
              as anything that includes a minus '-' sign or starts with a number).
        """
        super(MDP, self).__init__(
            **kwargs
        )  # can use kwargs to set dict! (but no sanity checks!)

        self.autoconvert = autoconvert

        if filename is not None:
            self._init_filename(filename)
            self.read(filename)

    def __eq__(self, other):
        """
        __eq__ inherited from FileUtils needs to be overridden if new attributes (autoconvert in
        this case) are assigned to the instance of the subclass (MDP in our case).
        See `this post by LGTM <https://lgtm.com/rules/9990086/>`_ for more details.
        """
        if not isinstance(other, MDP):
            return False
        return FileUtils.__eq__(self, other) and self.autoconvert == other.autoconvert

    def _transform(self, value):
        if self.autoconvert:
            return utils.autoconvert(value)
        else:
            return value.rstrip()

    def read(self, filename=None):
        """Read and parse mdp file *filename*."""
        self._init_filename(filename)

        def BLANK(i):
            return "B{0:04d}".format(i)

        def COMMENT(i):
            return "C{0:04d}".format(i)

        data = odict()
        iblank = icomment = 0
        with open(self.real_filename) as mdp:
            for line in mdp:
                line = line.strip()
                if len(line) == 0:
                    iblank += 1
                    data[BLANK(iblank)] = ""
                    continue
                m = self.COMMENT.match(line)
                if m:
                    icomment += 1
                    data[COMMENT(icomment)] = m.group("value")
                    continue
                # parameter
                m = self.PARAMETER.match(line)
                if m:
                    # check for comments after parameter?? -- currently discarded
                    parameter = m.group("parameter")
                    value = self._transform(m.group("value"))
                    data[parameter] = value
                else:
                    errmsg = "{filename!r}: unknown line in mdp file, {line!r}".format(
                        **vars()
                    )
                    self.logger.error(errmsg)
                    raise ParseError(errmsg)

        super(MDP, self).update(data)

    def write(self, filename=None, skipempty=False):
        """Write mdp file to *filename*.

        Parameters
        ----------
        filename : str
            Output mdp file; default is the filename the mdp was read from. If the filename
            is not supplied, the function will overwrite the file that the mdp was read from.
        skipempty : bool
            ``True`` removes any parameter lines from output that contain empty values [``False``]
        """
        # The line 'if skipempty and (v == "" or v is None):' below could possibly incur FutureWarning
        warnings.simplefilter(action='ignore', category=FutureWarning)

        with open(self.filename(filename, ext="mdp"), "w") as mdp:
            for k, v in self.items():
                if k[0] == "B":  # blank line
                    mdp.write("\n")
                elif k[0] == "C":  # comment
                    mdp.write("; {v!s}\n".format(**vars()))
                else:  # parameter = value
                    if skipempty and (v == "" or v is None):
                        continue
                    if isinstance(v, six.string_types) or not hasattr(v, "__iter__"):
                        mdp.write("{k!s} = {v!s}\n".format(**vars()))
                    else:
                        mdp.write("{} = {}\n".format(k, " ".join(map(str, v))))



####################################################################
#                                                                  #
#    ensemble_md,                                                  #
#    a python package for running GROMACS simulation ensembles     #
#                                                                  #
#    Written by Wei-Tse Hsu <wehs7661@colorado.edu>                #
#    Copyright (c) 2022 University of Colorado Boulder             #
#                                                                  #
####################################################################
"""
Unit tests for the module ensemble_EXE.py.
"""
import os
import sys
import yaml
import copy
import random
import shutil
import pytest
import numpy as np
import ensemble_md
import gmxapi as gmx
from mpi4py import MPI
from ensemble_md.utils import gmx_parser
from ensemble_md.ensemble_EXE import EnsembleEXE
from ensemble_md.utils.exceptions import ParameterError

current_path = os.path.dirname(os.path.abspath(__file__))
input_path = os.path.join(current_path, "data")


@pytest.fixture
def params_dict():
    """
    Generates a dictionary containing the required EEXE parameters.
    """
    EEXE_dict = {
        'parallel': False,
        'gro': 'ensemble_md/tests/data/sys.gro',
        'top': 'ensemble_md/tests/data/sys.top',
        'mdp': 'ensemble_md/tests/data/expanded.mdp',
        'n_sim': 4,
        'n_iter': 10,
        's': 1,
    }
    yield EEXE_dict

    # Remove the file after the unit test is done.
    if os.path.isfile('params.yaml') is True:
        os.remove('params.yaml')


def get_EEXE_instance(input_dict, yml_file='params.yaml'):
    """
    Saves a dictionary as a yaml file and use it to instantiate the EnsembleEXE class.
    """
    with open(yml_file, 'w') as f:
        yaml.dump(input_dict, f)
    EEXE = EnsembleEXE(yml_file)
    return EEXE


def check_param_error(EEXE_dict, param, match, wrong_val='cool', right_val=None):
    """
    Test if ParameterError is raised if a parameter is not well-defined.
    """
    EEXE_dict[param] = wrong_val
    with pytest.raises(ParameterError, match=match):
        get_EEXE_instance(EEXE_dict)
    EEXE_dict[param] = right_val  # so that EEXE_dict can be read without failing in later assertions

    return EEXE_dict


class Test_EnsembleEXE:
    def test_init(self, params_dict):
        EEXE = get_EEXE_instance(params_dict)
        assert EEXE.yaml == 'params.yaml'

    def test_set_params_error(self, params_dict):
        # 1. Required parameters
        del params_dict['parallel']
        with pytest.raises(ParameterError, match="Required parameter 'parallel' not specified in params.yaml."):
            get_EEXE_instance(params_dict)
        params_dict['parallel'] = False  # so params_dict can be read without failing in the assertions below

        # 2. Available options
        check_param_error(params_dict, 'proposal', "The specified proposal scheme is not available. Available options include 'single', 'neighboring', 'exhaustive', and 'multiple'.", 'cool', 'multiple')  # set as multiple for later tests for n_ex # noqa: E501
        check_param_error(params_dict, 'acceptance', "The specified acceptance scheme is not available. Available options include 'same-state', 'metropolis', and 'metropolis-eq'.")  # noqa: E501
        check_param_error(params_dict, 'w_scheme', "The specified weight combining scheme is not available. Available options include None, 'mean', 'geo-mean'/'geo_mean' and 'g-diff/g_diff'.")  # noqa: E501
        check_param_error(params_dict, 'df_method', "The specified free energy estimator is not available. Available options include 'TI', 'BAR', and 'MBAR'.")  # noqa: E501
        check_param_error(params_dict, 'err_method', "The specified method for error estimation is not available. Available options include 'propagate', and 'bootstrap'.")  # noqa: E501

        # 3. Integer parameters
        check_param_error(params_dict, 'nst_sim', "The parameter 'nst_sim' should be an integer.")
        check_param_error(params_dict, 'n_ex', "The parameter 'n_ex' should be an integer.")
        check_param_error(params_dict, 'seed', "The parameter 'seed' should be an integer.")
        check_param_error(params_dict, 'n_sim', "The parameter 'n_sim' should be an integer.", 4.1, 4)

        # 4. Positive parameters
        check_param_error(params_dict, 'nst_sim', "The parameter 'nst_sim' should be positive.", -5, 500)
        check_param_error(params_dict, 'n_iter', "The parameter 'n_iter' should be positive.", 0, 10)

        # 5. Non-negative parameters
        check_param_error(params_dict, 'n_ex', "The parameter 'n_ex' should be non-negative.", -1)
        check_param_error(params_dict, 'N_cutoff', "The parameter 'N_cutoff' should be non-negative unless no histogram correction is needed, i.e. N_cutoff = -1.", -5)  # noqa: E501

        # 6. String parameters
        check_param_error(params_dict, 'mdp', "The parameter 'mdp' should be a string.", 3, 'ensemble_md/tests/data/expanded.mdp')  # noqa: E501

        # 7. Boolean parameters
        check_param_error(params_dict, 'parallel', "The parameter 'parallel' should be a boolean variable.", 3, False)

        # 8. nstlog > nst_sim
        mdp = gmx_parser.MDP(os.path.join(input_path, "expanded.mdp"))  # A perfect mdp file
        mdp['nstlog'] = 200
        mdp.write(os.path.join(input_path, "expanded_test.mdp"))
        params_dict['mdp'] = 'ensemble_md/tests/data/expanded_test.mdp'
        params_dict['nst_sim'] = 100
        with pytest.raises(ParameterError, match='The parameter "nstlog" should be equal to or smaller than "nst_sim" specified in the YAML file so that the sampling information can be parsed.'):  # noqa: E501
            get_EEXE_instance(params_dict)
        params_dict['nst_sim'] = 500
        os.remove(os.path.join(input_path, "expanded_test.mdp"))

        # 9. n_sub < 1
        params_dict['s'] = 5
        params_dict['mdp'] = 'ensemble_md/tests/data/expanded.mdp'
        # Note that the parentheses are special characters that need to be escaped in regular expressions
        with pytest.raises(ParameterError, match=r'There must be at least two states for each replica \(current value: -6\). The current specified configuration \(n_tot=9, n_sim=4, s=5\) does not work for EEXE.'):  # noqa: E501
            get_EEXE_instance(params_dict)
        params_dict['s'] = 1

    def test_set_params_warnings(self, params_dict):
        # 1. Non-recognizable parameter in the YAML file
        params_dict['cool'] = 10
        EEXE = get_EEXE_instance(params_dict)
        warning = 'Warning: Parameter "cool" specified in the input YAML file is not recognizable.'
        assert warning in EEXE.warnings

        # 2. Warnings related to the mdp file
        mdp = gmx_parser.MDP(os.path.join(input_path, "expanded.mdp"))  # A perfect mdp file
        mdp['lmc_seed'] = 1000
        mdp['gen_seed'] = 1000
        mdp['symmetrized_transition_matrix'] = 'yes'
        mdp.write(os.path.join(input_path, "expanded_test.mdp"))
        params_dict['mdp'] = 'ensemble_md/tests/data/expanded_test.mdp'
        EEXE = get_EEXE_instance(params_dict)
        warning_1 = 'Warning: We recommend setting lmc_seed as -1 so the random seed is different for each iteration.'
        warning_2 = 'Warning: We recommend setting gen_seed as -1 so the random seed is different for each iteration.'
        warning_3 = 'Warning: We recommend setting symmetrized-transition-matrix to no instead of yes.'
        assert warning_1 in EEXE.warnings
        assert warning_2 in EEXE.warnings
        assert warning_3 in EEXE.warnings

        os.remove(os.path.join(input_path, "expanded_test.mdp"))

    def test_set_params(self, params_dict):
        # 0. Get an EEXE instance to test
        EEXE = get_EEXE_instance(params_dict)

        # 1. Check the required EEXE parameters
        assert EEXE.gro == "ensemble_md/tests/data/sys.gro"
        assert EEXE.top == "ensemble_md/tests/data/sys.top"
        assert EEXE.mdp == "ensemble_md/tests/data/expanded.mdp"
        assert EEXE.parallel is False
        assert EEXE.n_sim == 4
        assert EEXE.n_iter == 10
        assert EEXE.s == 1

        # 2. Check the default values of the parameters not specified in params.yaml
        assert EEXE.proposal == "exhaustive"
        assert EEXE.acceptance == "metropolis"
        assert EEXE.w_scheme is None
        assert EEXE.N_cutoff == 1000
        assert EEXE.n_ex == 'N^3'
        assert EEXE.verbose is True
        assert EEXE.runtime_args is None
        assert EEXE.n_ckpt == 100
        assert EEXE.msm is False
        assert EEXE.free_energy is False
        assert EEXE.df_spacing == 1
        assert EEXE.df_method == 'MBAR'
        assert EEXE.err_method == 'propagate'
        assert EEXE.n_bootstrap == 50
        assert EEXE.seed is None

        # 3. Check the MDP parameters
        assert EEXE.nsteps == 500
        assert EEXE.dt == 0.002
        assert EEXE.temp == 298
        assert EEXE.nst_sim == 500
        assert EEXE.fixed_weights is False

        # 4. Checked the derived parameters
        # Note that lambda_dict and lambda_ranges will also be tested in test_map_lambda2state.
        k = 1.380649e-23
        NA = 6.0221408e23
        assert EEXE.kT == k * NA * 298 / 1000
        assert EEXE.lambda_types == ['coul_lambdas', 'vdw_lambdas']
        assert EEXE.n_tot == 9
        assert EEXE.n_sub == 6
        assert EEXE.state_ranges == [
            [0, 1, 2, 3, 4, 5],
            [1, 2, 3, 4, 5, 6],
            [2, 3, 4, 5, 6, 7],
            [3, 4, 5, 6, 7, 8], ]
        assert EEXE.equil == [-1, -1, -1, -1]  # will be zeros right after the first iteration if the weights are fixed
        assert EEXE.n_rejected == 0
        assert EEXE.n_swap_attempts == 0
        assert EEXE.rep_trajs == [[0], [1], [2], [3]]

        params_dict['df_method'] = 'MBAR'
        EEXE = get_EEXE_instance(params_dict)
        assert EEXE.get_u_nk is True
        assert EEXE.get_dHdl is False

        params_dict['df_method'] = 'BAR'
        EEXE = get_EEXE_instance(params_dict)
        assert EEXE.get_u_nk is False
        assert EEXE.get_dHdl is True

        params_dict['runtime_args'] = {'-nt': 16, 'ntomp': 8}
        EEXE = get_EEXE_instance(params_dict)
        assert EEXE.runtime_args == {'-nt': 16, 'ntomp': 8}

        assert EEXE.reformatted_mdp is False

    def test_set_params_edge_cases(self, params_dict):
        # In the previous unit tests, we have tested the case where fixed_weights is False. Here we test the others.
        mdp = gmx_parser.MDP(os.path.join(input_path, "expanded.mdp"))

        # 1. wl_scale is not specified
        del mdp['wl_scale']
        mdp.write(os.path.join(input_path, "expanded_test.mdp"))
        params_dict['mdp'] = 'ensemble_md/tests/data/expanded_test.mdp'
        EEXE = get_EEXE_instance(params_dict)
        assert EEXE.fixed_weights is True

        # 2. wl_scale is not empty
        mdp['wl_scale'] = ''  # It will become an empty array array([]) after read by gmx_parser.abs(x)
        mdp.write(os.path.join(input_path, "expanded_test.mdp"))
        EEXE = get_EEXE_instance(params_dict)
        assert EEXE.fixed_weights is True

    def test_reformat_MDP(self, params_dict):
        # Note that the function reformat_MDP is called in set_params
        mdp = gmx_parser.MDP(os.path.join(input_path, "expanded.mdp"))
        mdp['cool-stuff'] = 10
        mdp.write(os.path.join(input_path, "expanded_test.mdp"))
        params_dict['mdp'] = 'ensemble_md/tests/data/expanded_test.mdp'
        EEXE = get_EEXE_instance(params_dict)  # this calss reformat_MDP internally
        assert EEXE.reformatted_mdp is True
        assert EEXE.template['cool_stuff'] == 10
        assert os.path.isfile(os.path.join(input_path, "expanded_test_backup.mdp"))

        os.remove(os.path.join(input_path, "expanded_test.mdp"))
        os.remove(os.path.join(input_path, "expanded_test_backup.mdp"))

    def test_map_lambda2state(self, params_dict):
        # Note that the function map_lambda2state is called in set_params
        EEXE = get_EEXE_instance(params_dict)
        assert EEXE.lambda_dict == {
            (0, 0): 0,
            (0.25, 0): 1,
            (0.5, 0): 2,
            (0.75, 0): 3,
            (1, 0): 4,
            (1, 0.25): 5,
            (1, 0.5): 6,
            (1, 0.75): 7,
            (1, 1): 8}
        assert EEXE.lambda_ranges == [
            [(0.0, 0.0), (0.25, 0.0), (0.5, 0.0), (0.75, 0.0), (1.0, 0.0), (1.0, 0.25)],
            [(0.25, 0.0), (0.5, 0.0), (0.75, 0.0), (1.0, 0.0), (1.0, 0.25), (1.0, 0.5)],
            [(0.5, 0.0), (0.75, 0.0), (1.0, 0.0), (1.0, 0.25), (1.0, 0.5), (1.0, 0.75)],
            [(0.75, 0.0), (1.0, 0.0), (1.0, 0.25), (1.0, 0.5), (1.0, 0.75), (1.0, 1.0)], ]

        # Here we test another combinations: only 'fep_lambdas'
        mdp = gmx_parser.MDP(os.path.join(input_path, "expanded.mdp"))
        del mdp['coul_lambdas']
        del mdp['vdw_lambdas']
        mdp['fep_lambdas'] = '0.0 0.2 0.4 0.6 0.8 1.0'
        mdp.write(os.path.join(input_path, "expanded_test.mdp"))
        params_dict['mdp'] = 'ensemble_md/tests/data/expanded_test.mdp'
        params_dict['n_sim'] = 3
        EEXE = get_EEXE_instance(params_dict)
        assert EEXE.lambda_types == ['fep_lambdas']
        assert EEXE.lambda_dict == {
            (0.0,): 0, (0.2,): 1, (0.4,): 2, (0.6,): 3, (0.8,): 4, (1.0,): 5}
        assert EEXE.lambda_ranges == [
            [(0.0, ), (0.2,), (0.4,), (0.6,)], [(0.2,), (0.4,), (0.6,), (0.8,)], [(0.4,), (0.6,), (0.8,), (1.0,)]]

        # Still another combination: fep_lambdas, coul_lambdas, vdw_lambdas, restraint_lambdas
        mdp['coul_lambdas'] = '0.0 0.5 0.8 1.0 1.0 1.0'
        mdp['vdw_lambdas'] = '0.0 0.0 0.0 0.0 0.5 1.0'
        mdp['restraint_lambdas'] = '0.0 0.2 0.4 0.6 0.8 1.0'
        mdp.write(os.path.join(input_path, "expanded_test.mdp"))
        EEXE = get_EEXE_instance(params_dict)
        assert EEXE.lambda_types == ['fep_lambdas', 'coul_lambdas', 'vdw_lambdas', 'restraint_lambdas']
        assert EEXE.lambda_dict == {
            (0.0, 0.0, 0.0, 0.0): 0,
            (0.2, 0.5, 0.0, 0.2): 1,
            (0.4, 0.8, 0.0, 0.4): 2,
            (0.6, 1.0, 0.0, 0.6): 3,
            (0.8, 1.0, 0.5, 0.8): 4,
            (1.0, 1.0, 1.0, 1.0): 5}

        assert EEXE.lambda_ranges == [
            [(0.0, 0.0, 0.0, 0.0), (0.2, 0.5, 0.0, 0.2), (0.4, 0.8, 0.0, 0.4), (0.6, 1.0, 0.0, 0.6)],
            [(0.2, 0.5, 0.0, 0.2), (0.4, 0.8, 0.0, 0.4), (0.6, 1.0, 0.0, 0.6), (0.8, 1.0, 0.5, 0.8)],
            [(0.4, 0.8, 0.0, 0.4), (0.6, 1.0, 0.0, 0.6), (0.8, 1.0, 0.5, 0.8), (1.0, 1.0, 1.0, 1.0)]]

        os.remove(os.path.join(input_path, "expanded_test.mdp"))

    def test_print_params(self, capfd, params_dict):
        # capfd is a fixture in pytest for testing STDOUT
        EEXE = get_EEXE_instance(params_dict)
        EEXE.print_params()
        out_1, err = capfd.readouterr()
        L = ""
        L += "Important parameters of EXEE\n============================\n"
        L += f"Python version: {sys.version}\ngmxapi version: {gmx.__version__}\nensemble_md version: {ensemble_md.__version__}\n"  # noqa: E501
        L += "Simulation inputs: ensemble_md/tests/data/sys.gro, ensemble_md/tests/data/sys.top, ensemble_md/tests/data/expanded.mdp\n"  # noqa: E501
        L += "Verbose log file: True\nWhether the replicas run in parallel: False\n"
        L += "Proposal scheme: exhaustive\n"
        L += "Acceptance scheme for swapping simulations: metropolis\nScheme for combining weights: None\n"
        L += "Histogram cutoff: 1000\nNumber of replicas: 4\nNumber of iterations: 10\n"
        L += "Number of attempted swaps in one exchange interval: N^3\n"
        L += "Length of each replica: 1.0 ps\nFrequency for checkpointing: 100 iterations\n"
        L += "Total number of states: 9\n"
        L += "Additional runtime arguments: None\n"
        L += "Alchemical ranges of each replica in EEXE:\n  - Replica 0: States [0, 1, 2, 3, 4, 5]\n"
        L += "  - Replica 1: States [1, 2, 3, 4, 5, 6]\n  - Replica 2: States [2, 3, 4, 5, 6, 7]\n"
        L += "  - Replica 3: States [3, 4, 5, 6, 7, 8]\n"
        assert out_1 == L

        EEXE.reformatted_mdp = True  # Just to test the case where EEXE.reformatted_mdp is True
        EEXE.print_params(params_analysis=True)
        out_2, err = capfd.readouterr()
        L += "\nWhether to build Markov state models and perform relevant analysis: False\n"
        L += "Whether to perform free energy calculations: False\n"
        L += "The step to used in subsampling the DHDL data in free energy calculations, if any: 1\n"
        L += "The chosen free energy estimator for free energy calculations, if any: MBAR\n"
        L += "The method for estimating the uncertainty of free energies in free energy calculations, if any: propagate\n"  # noqa: E501
        L += "The number of bootstrap iterations in the boostrapping method, if used: 50\n"
        L += "The random seed to use in bootstrapping, if used: None\n"
        L += "Note that the input MDP file has been reformatted by replacing hypens with underscores. The original mdp file has been renamed as *backup.mdp.\n"  # noqa: E501
        assert out_2 == L

    def test_initialize_MDP(self, params_dict):
        EEXE = get_EEXE_instance(params_dict)
        MDP = EEXE.initialize_MDP(2)  # the third replica
        assert MDP["nsteps"] == 500
        assert all(
            [
                a == b
                for a, b in zip(
                    MDP["vdw_lambdas"], [0.00, 0.00, 0.00, 0.25, 0.50, 0.75]
                )
            ]
        )
        assert all(
            [
                a == b
                for a, b in zip(
                    MDP["coul_lambdas"], [0.50, 0.75, 1.00, 1.00, 1.00, 1.00]
                )
            ]
        )
        assert all(
            [a == b for a, b in zip(MDP["init_lambda_weights"], [0, 0, 0, 0, 0, 0])]
        )

    def test_update_MDP(self, params_dict):
        new_template = "ensemble_md/tests/data/expanded.mdp"
        iter_idx = 3
        states = [2, 5, 7, 4]
        wl_delta = [0.4, 0.32, 0.256, 0.32]
        weights = [
            [0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0],
            [3.48, 2.78, 3.21, 4.56, 8.79, 0.48],
            [8.45, 0.52, 3.69, 2.43, 4.56, 6.73], ]

        EEXE = get_EEXE_instance(params_dict)
        EEXE.equil = [-1, 1, 0, -1]  # i.e. the 3rd replica will use fixed weights in the next iteration
        MDP_1 = EEXE.update_MDP(
            new_template, 2, iter_idx, states, wl_delta, weights)  # third replica
        MDP_2 = EEXE.update_MDP(
            new_template, 3, iter_idx, states, wl_delta, weights)  # fourth replica

        assert MDP_1["tinit"] == MDP_2["tinit"] == 3
        assert MDP_1["nsteps"] == MDP_2["nsteps"] == 500
        assert MDP_1["init_lambda_state"] == 5
        assert MDP_2["init_lambda_state"] == 1
        assert (MDP_1["init_wl_delta"] == MDP_1["wl_scale"] == MDP_1["wl_ratio"] == "")
        assert (MDP_1["lmc_weights_equil"] == MDP_1["weight_equil_wl_delta"] == "")
        assert MDP_2["init_wl_delta"] == 0.32
        assert all(
            [
                a == b
                for a, b in zip(
                    MDP_1["init_lambda_weights"], [3.48, 2.78, 3.21, 4.56, 8.79, 0.48]
                )
            ]
        )
        assert all(
            [
                a == b
                for a, b in zip(
                    MDP_2["init_lambda_weights"], [8.45, 0.52, 3.69, 2.43, 4.56, 6.73]
                )
            ]
        )

    def test_extract_final_dhdl_info(self, params_dict):
        EEXE = get_EEXE_instance(params_dict)
        dhdl_files = [
            os.path.join(input_path, f"dhdl/dhdl_{i}.xvg") for i in range(EEXE.n_sim)
        ]
        states = EEXE.extract_final_dhdl_info(dhdl_files)
        assert states == [5, 2, 2, 8]

    def test_extract_final_log_info(self, params_dict):
        EEXE = get_EEXE_instance(params_dict)
        log_files = [
            os.path.join(input_path, f"log/EXE_{i}.log") for i in range(EEXE.n_sim)
        ]
        wl_delta, weights, counts = EEXE.extract_final_log_info(log_files)
        assert wl_delta == [0.4, 0.5, 0.5, 0.5]
        assert np.allclose(weights, [
            [0, 1.03101, 2.55736, 3.63808, 4.47220, 6.13408],
            [0, 1.22635, 2.30707, 2.44120, 4.10308, 6.03106],
            [0, 0.66431, 1.25475, 0.24443, 0.59472, 0.70726],
            [0, 0.09620, 1.59937, -4.31679, -22.89436, -28.08701], ])
        assert counts == [
            [4, 11, 9, 9, 11, 6],
            [9, 8, 8, 11, 7, 7],
            [3, 1, 1, 9, 15, 21],
            [0, 0, 0, 1, 18, 31], ]
        assert EEXE.equil == [-1, -1, -1, -1]

    def test_identify_swappable_pairs(self, params_dict):
        EEXE = get_EEXE_instance(params_dict)
        EEXE.state_ranges = [list(range(i, i + 5)) for i in range(EEXE.n_sim)]  # 5 states per replica
        states = [4, 2, 2, 7]   # This would lead to the swappables: [(0, 1), (0, 2), (1, 2)] in the standard case

        # Case 1: Any case that is not neighboring swap has the same definition for the swappable pairs
        swappables_1 = EEXE.identify_swappable_pairs(states, EEXE.state_ranges)
        assert swappables_1 == [(0, 1), (0, 2), (1, 2)]

        # Case 2: Neighboring exchange
        EEXE.proposal = 'neighboring'
        swappables_2 = EEXE.identify_swappable_pairs(states, EEXE.state_ranges)
        assert swappables_2 == [(0, 1), (1, 2)]

    def test_propose_swap(self, params_dict):
        random.seed(0)
        EEXE = get_EEXE_instance(params_dict)
        swap_1 = EEXE.propose_swap([])
        swap_2 = EEXE.propose_swap([(0, 1), (0, 2), (1, 2)])
        assert swap_1 == []
        assert swap_2 == (1, 2)

    def test_get_swapping_pattern(self, params_dict):
        # weights are obtained from the log files in data/log, where the last states are 5, 2, 2, 8 (global indices)
        # state_ranges are: 0-5, 1-6, ..., 3-8
        weights = [
            [0, 1.03101, 2.55736, 3.63808, 4.47220, 6.13408],
            [0, 1.22635, 2.30707, 2.44120, 4.10308, 6.03106],
            [0, 0.66431, 1.25475, 1.24443, 0.59472, 0.70726],   # the 4th prob was ajusted (from 0.24443) to tweak prob_acc  # noqa: E501
            [0, 0.09620, 1.59937, -4.31679, -22.89436, -28.08701]]
        dhdl_files = [os.path.join(input_path, f"dhdl/dhdl_{i}.xvg") for i in range(4)]

        # Case 1: Empty swap list
        EEXE = get_EEXE_instance(params_dict)
        EEXE.verbose = False
        states = [0, 6, 7, 8]  # No swappable pairs
        w, f = copy.deepcopy(weights), copy.deepcopy(dhdl_files)
        pattern_1 = EEXE.get_swapping_pattern(f, states, w)
        assert EEXE.n_swap_attempts == 1
        assert EEXE.n_rejected == 1
        assert pattern_1 == [0, 1, 2, 3]

        # Case 2: Single swap (proposal = 'single')
        random.seed(0)
        EEXE = get_EEXE_instance(params_dict)
        EEXE.verbose = True
        EEXE.proposal = 'single'  # n_ex will be set to 1 automatically.
        states = [5, 2, 2, 8]  # swappable pairs: [(0, 1), (0, 2), (1, 2)], swap = (1, 2), accept
        w, f = copy.deepcopy(weights), copy.deepcopy(dhdl_files)
        pattern_2 = EEXE.get_swapping_pattern(f, states, w)
        assert EEXE.n_swap_attempts == 1
        assert EEXE.n_rejected == 0
        assert pattern_2 == [0, 2, 1, 3]

        # Case 3: Neighboring swap
        random.seed(0)
        EEXE = get_EEXE_instance(params_dict)
        EEXE.proposal = 'neighboring'  # n_ex will be set to 1 automatically.
        states = [5, 2, 2, 8]  # swappable pairs: [(0, 1), (1, 2)], swap = (1, 2), accept
        w, f = copy.deepcopy(weights), copy.deepcopy(dhdl_files)
        pattern_3 = EEXE.get_swapping_pattern(f, states, w)
        assert EEXE.n_swap_attempts == 1
        assert EEXE.n_rejected == 0
        assert pattern_3 == [0, 2, 1, 3]

        # Case 4-1: Exhaustive swaps that end up in a single swap
        random.seed(0)
        EEXE = get_EEXE_instance(params_dict)
        EEXE.proposal = 'exhaustive'
        states = [5, 2, 2, 8]  # swappable pairs: [(0, 1), (0, 2), (1, 2)], swap = (1, 2), accept
        w, f = copy.deepcopy(weights), copy.deepcopy(dhdl_files)
        pattern_4_1 = EEXE.get_swapping_pattern(f, states, w)
        assert EEXE.n_swap_attempts == 1
        assert EEXE.n_rejected == 0
        assert pattern_4_1 == [0, 2, 1, 3]

        # Case 4-2: Exhaustive swaps that involve multiple attempted swaps
        random.seed(0)
        EEXE = get_EEXE_instance(params_dict)
        EEXE.proposal = 'exhaustive'
        states = [4, 2, 4, 3]  # swappable pairs: [(0, 1), (0, 2), (0, 3), (1, 2), (2, 3)]; swap 1: (2, 3), accepted; swap 2: (0, 1), accept  # noqa: E501
        w, f = copy.deepcopy(weights), copy.deepcopy(dhdl_files)
        pattern_4_2 = EEXE.get_swapping_pattern(f, states, w)
        assert EEXE.n_swap_attempts == 2   # \Delta is negative for both swaps -> both accepted
        assert EEXE.n_rejected == 0
        assert pattern_4_2 == [1, 0, 3, 2]

        # Case 5-1: Multiple swaps (proposal = 'multiple', n_ex = 5)
        print('test 5-1')
        random.seed(0)
        EEXE = get_EEXE_instance(params_dict)
        EEXE.n_ex = 5
        EEXE.proposal = 'multiple'
        states = [3, 1, 4, 6]  # swappable pairs: [(0, 1), (0, 2), (1, 2)], first swap = (1, 2), accept
        w, f = copy.deepcopy(weights), copy.deepcopy(dhdl_files)
        pattern_5_1 = EEXE.get_swapping_pattern(f, states, w)
        assert EEXE.n_swap_attempts == 5
        assert EEXE.n_rejected == 4
        assert pattern_5_1 == [2, 1, 0, 3]

        # Case 5-2: Multiple swaps but with only one swappable pair (proposal = 'multiple')
        # This is specifically for testing n_swap_attempts in the case where n_ex reduces to 1.
        random.seed(0)
        EEXE = get_EEXE_instance(params_dict)
        states = [0, 2, 3, 8]  # The only swappable pair is [(1, 2)] --> accept
        w, f = copy.deepcopy(weights), copy.deepcopy(dhdl_files)
        pattern_5_2 = EEXE.get_swapping_pattern(f, states, w)
        assert EEXE.n_swap_attempts == 1  # since there is only 1 swappable pair
        assert EEXE.n_rejected == 0
        assert pattern_5_2 == [0, 2, 1, 3]

    def test_calc_prob_acc(self, capfd, params_dict):
        EEXE = get_EEXE_instance(params_dict)
        # EEXE.state_ranges = [[0, 1, 2, 3, 4, 5], [1, 2, 3, 4, 5, 6], ..., [3, 4, 5, 6, 7, 8]]
        states = [5, 2, 2, 8]
        shifts = [0, 1, 2, 3]
        weights = [
            [0, 1.03101, 2.55736, 3.63808, 4.47220, 6.13408],
            [0, 1.22635, 2.30707, 2.44120, 4.10308, 6.03106],
            [0, 0.66431, 1.25475, 0.24443, 0.59472, 0.70726],
            [0, 0.09620, 1.59937, -4.31679, -22.89436, -28.08701]]
        dhdl_files = [os.path.join(input_path, f"dhdl/dhdl_{i}.xvg") for i in range(4)]

        # Test 1: Same-state swapping (True)
        swap = (1, 2)
        EEXE.acceptance = "same_state"
        prob_acc_1 = EEXE.calc_prob_acc(swap, dhdl_files, states, shifts, weights)
        assert prob_acc_1 == 1

        # Test 2: Same-state swapping (False)
        swap = (0, 2)
        prob_acc_2 = EEXE.calc_prob_acc(swap, dhdl_files, states, shifts, weights)
        assert prob_acc_2 == 0

        # Test 3: Metropolis-eq
        swap = (0, 2)
        EEXE.acceptance = "metropolis-eq"
        prob_acc_3 = EEXE.calc_prob_acc(swap, dhdl_files, states, shifts, weights)
        assert prob_acc_3 == 1    # Delta U = (-9.1366697 + 4.9963939)/2.478956208925815 ~ -1.67 kT

        # Test 4: Metropolis
        swap = (0, 2)
        EEXE.acceptance = "metropolis"
        prob_acc_4 = EEXE.calc_prob_acc(swap, dhdl_files, states, shifts, weights)
        out, err = capfd.readouterr()
        # dH ~-1.67 kT as calculated above, dg = (2.55736 - 6.13408) + (0.24443 - 0) ~ -3.33229 kT
        # dU - dg ~ 1.66212 kT, so p_acc ~ 0.189 ...
        assert prob_acc_4 == pytest.approx(0.18989559074633955)   # check this number again
        assert 'U^i_n - U^i_m = -3.69 kT, U^j_m - U^j_n = 2.02 kT, Total dU: -1.67 kT' in out
        assert 'g^i_n - g^i_m = -3.58 kT, g^j_m - g^j_n = 0.24 kT, Total dg: -3.33 kT' in out

        # Test 5: Additional test: the case where one swap of (0, 2) has been accepted already.
        # Given another same swap of (0, 2), dU and dg in this case should be equal opposites of the oens in Case 4.
        # And the acceptance ratio should be 1.
        swap = (0, 2)
        states = [2, 2, 5, 8]
        prob_acc_5 = EEXE.calc_prob_acc(swap, dhdl_files, states, shifts, weights)
        out, err = capfd.readouterr()
        assert prob_acc_5 == 1
        assert 'U^i_n - U^i_m = 3.69 kT, U^j_m - U^j_n = -2.02 kT, Total dU: 1.67 kT' in out
        assert 'g^i_n - g^i_m = 3.58 kT, g^j_m - g^j_n = -0.24 kT, Total dg: 3.33 kT' in out

    def test_accept_or_reject(self, params_dict):
        EEXE = get_EEXE_instance(params_dict)
        random.seed(0)
        swap_bool_1 = EEXE.accept_or_reject(0)
        swap_bool_2 = EEXE.accept_or_reject(0.8)  # rand = 0.844
        swap_bool_3 = EEXE.accept_or_reject(0.8)  # rand = 0.758

        assert EEXE.n_swap_attempts == 0  # since we didn't use get_swapping_pattern
        assert EEXE.n_rejected == 2
        assert swap_bool_1 is False
        assert swap_bool_2 is False
        assert swap_bool_3 is True

    def test_historgam_correction(self, params_dict):
        EEXE = get_EEXE_instance(params_dict)
        # Case 1: No histogram correction
        EEXE.N_cutoff = -1
        weights_1 = [[0, 10.304, 20.073, 29.364]]
        counts_1 = [[31415, 45701, 55457, 59557]]
        weights_1 = EEXE.histogram_correction(weights_1, counts_1)
        assert weights_1 == [[0, 10.304, 20.073, 29.364]]

        # Case 2: Perform histogram correction (N_cutoff reached)
        EEXE.N_cutoff = 5000
        EEXE.verbose = False  # just to increase code coverage
        weights_1 = EEXE.histogram_correction(weights_1, counts_1)
        assert np.allclose(weights_1, [
            [
                0,
                10.304 + np.log(31415 / 45701),
                20.073 + np.log(45701 / 55457),
                29.364 + np.log(55457 / 59557),
            ]
        ])  # noqa: E501

        # Case 3: Perform histogram correction (N_cutoff not reached by both N_k and N_{k-1})
        EEXE.verbose = True
        weights_2 = [[0, 10.304, 20.073, 29.364]]
        counts_2 = [[3141, 4570, 5545, 5955]]
        weights_2 = EEXE.histogram_correction(weights_2, counts_2)
        assert np.allclose(weights_2, [[0, 10.304, 20.073, 29.364 + np.log(5545 / 5955)]])

    def test_combine_weights(self, params_dict):
        EEXE = get_EEXE_instance(params_dict)
        EEXE.n_tot = 6
        EEXE.n_sub = 4
        EEXE.s = 1
        EEXE.n_sim = 3
        EEXE.state_ranges = [[0, 1, 2, 3], [1, 2, 3, 4], [2, 3, 4, 5]]
        weights = [[0, 2.1, 4.0, 3.7], [0, 1.7, 1.2, 2.6], [0, -0.4, 0.9, 1.9]]

        # Method: None
        w1, g_vec_1 = EEXE.combine_weights(weights, method=None)
        assert np.allclose(w1, weights)
        assert g_vec_1 is None

        # Method: mean
        EEXE.verbose = False  # just to increase code coverage
        w2, g_vec_2 = EEXE.combine_weights(weights, method='mean')
        assert np.allclose(w2, [
            [0.0, 2.20097, 3.99803, 3.59516],
            [0.0, 1.79706, 1.39419, 2.69607],
            [0.0, -0.40286, 0.89901, 1.88303]])
        assert np.allclose(list(g_vec_2), [0.0, 2.200968785917372, 3.9980269151210854, 3.5951633659351256, 4.897041830662871, 5.881054277773005])  # noqa: E501

        # Method: geo-mean
        EEXE.verbose = True
        w3, g_vec_3 = EEXE.combine_weights(weights, method='geo-mean')
        assert np.allclose(w3, [
            [0.0, 2.2, 3.98889, 3.58889],
            [0.0, 1.78889, 1.38889, 2.68333],
            [0.0, -0.4, 0.89444, 1.87778]])
        assert np.allclose(list(g_vec_3), [0.0, 2.1999999999999997, 3.9888888888888885, 3.5888888888888886, 4.883333333333334, 5.866666666666667])  # noqa: E501

        # Method: g-diff
        w4, g_vec_4 = EEXE.combine_weights(weights, method='g-diff')
        assert np.allclose(w4, [
            [0, 2.1, 3.9, 3.5],
            [0, 1.8, 1.4, 2.75],
            [0, -0.4, 0.95, 1.95]])
        assert np.allclose(list(g_vec_4), [0, 2.1, 3.9, 3.5, 4.85, 5.85])

    def test_run_EEXE(self, params_dict):
        # We probably can only test serial EEXE
        rank = MPI.COMM_WORLD.Get_rank()
        params_dict['runtime_args'] = {'-nt': 1}
        EEXE = get_EEXE_instance(params_dict)
        if rank == 0:
            for i in range(EEXE.n_sim):
                os.mkdir(f'sim_{i}')
                os.mkdir(f'sim_{i}/iteration_0')
                MDP = EEXE.initialize_MDP(i)
                MDP.write(f'sim_{i}/iteration_0/expanded.mdp', skipempty=True)
                shutil.copy('ensemble_md/tests/data/sys.gro', f'sim_{i}/iteration_0/sys.gro')
                shutil.copy('ensemble_md/tests/data/sys.top', f'sim_{i}/iteration_0/sys.top')

        md = EEXE.run_EEXE(0)   # just test the first iteration is fine
        assert md.output.returncode.result() == [0, 0, 0, 0]

        if rank == 0:
            os.system('rm -r sim_*')
            os.system('rm -r gmxapi.commandline.cli*_i0*')



####################################################################
#                                                                  #
#    ensemble_md,                                                  #
#    a python package for running GROMACS simulation ensembles     #
#                                                                  #
#    Written by Wei-Tse Hsu <wehs7661@colorado.edu>                #
#    Copyright (c) 2022 University of Colorado Boulder             #
#                                                                  #
####################################################################
"""
The :obj:`.analyze_traj` module provides methods for analyzing trajectories in EEXE.
"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator

from alchemlyb.parsing.gmx import _get_headers as get_headers
from alchemlyb.parsing.gmx import _extract_dataframe as extract_dataframe
from ensemble_md.utils import utils


def extract_state_traj(dhdl):
    """
    Extracts the state-space trajectory from a GROMACS DHDL file.
    Note that the state indices here are local indices.

    Parameters
    ----------
    dhdl : str
        The filename of the GROMACS DHDL file to be parsed.

    Returns
    -------
    traj : list
        A list that represents that state-space trajectory
    """
    traj = list(extract_dataframe(dhdl, headers=get_headers(dhdl))['Thermodynamic state'])

    return traj


def stitch_trajs(files, rep_trajs, shifts=None, dhdl=True, col_idx=-1):
    """
    Stitches the state-space/CV-space trajectories for each configuration from DHDL files
    or PLUMED output files generated at different iterations.

    Parameters
    ----------
    files : list
        A list of lists of file names of GROMACS DHDL files or general GROMACS XVG files or PLUMED ouptput files.
        Specifically, :code:`files[i]` should be a list containing the files of interest from all iterations in
        replica :code:`i`.
    rep_trajs : list
        A list of lists that represents the replica space trajectories for each configuration. For example,
        :code:`rep_trajs[0] = [0, 2, 3, 0, 1, ...]` means that configuration 0 transitioned to replica 2, then
        3, 0, 1, in iterations 1, 2, 3, 4, ..., respectively.
    shifts : list
        A list of values for shifting the state indices for each replica. The length of the list
        should be equal to the number of replicas. This is only needed when :code:`dhdl=True`.
    dhdl : bool
        Whether the input files are GROMACS dhdl files, in which case trajectories of global alchemical indices
        will be generated. If :code:`dhdl=False`, the input files must be readable by `numpy.loadtxt` assuming that
        the start of a comment is indicated by either the :code:`#` or :code:`@` characters.
        Such files include any GROMACS XVG files or PLUMED output files (output by plumed driver, for instance).
        In this case, trajectories of the configurational collective variable of interest are generated.
        The default is :code:`True`.
    col_idx : int
        The index of the column to be extracted from the input files. This is only needed when :code:`dhdl=False`,
        By default, we extract the last column.

    Returns
    -------
    trajs : list
        A list that contains lists of state-space/CV-space trajectory (in global indices) for each configuration.
        For example, :code:`trajs[i]` is the state-space/CV-space trajectory of configuration :code:`i`.
    """
    n_configs = len(files)  # number of starting configurations
    n_iter = len(files[0])  # number of iterations per replica

    # First figure out which dhdl/plumed output files each configuration corresponds to
    # files_sorted[i] contains the dhdl/plumed output files for configuration i sorted based on iteration indices
    files_sorted = [[] for i in range(n_configs)]
    for i in range(n_configs):
        for j in range(n_iter):
            files_sorted[i].append(files[rep_trajs[i][j]][j])

    # Then, stitch the trajectories for each configuration
    trajs = [[] for i in range(n_configs)]  # for each configuration
    for i in range(n_configs):
        for j in range(n_iter):
            if j == 0:
                if dhdl:
                    traj = extract_state_traj(files_sorted[i][j])
                else:
                    traj = np.loadtxt(files_sorted[i][j], comments=['#', '@'])[:, -1]
            else:
                # Starting from the 2nd iteration, we get rid of the first time frame the first
                # frame of iteration n+1 the is the same as the last frame of iteration n
                if dhdl:
                    traj = extract_state_traj(files_sorted[i][j])[1:]
                else:
                    traj = np.loadtxt(files_sorted[i][j], comments=['#', '@'])[:, -1][1:]

            if dhdl:  # Trajectories of global alchemical indices will be generated.
                shift_idx = rep_trajs[i][j]
                traj = list(np.array(traj) + shifts[shift_idx])
            trajs[i].extend(traj)

    return trajs


def traj2transmtx(traj, N, normalize=True):
    """
    Computes the transition matrix given a trajectory. For example, if a state-space
    trajectory from a EXE or HREX simulation given, a state transition matrix is returned.
    If a trajectory showing transitions between replicas in a EEXE simulation is given,
    a replica transition matrix is returned.

    Parameters
    ---------
    traj : list
        A list of state indices showing the trajectory in the state space.
    N : int
        The size (N) of the expcted transition matrix (N by N).
    normalize : bool
        Whether to normalize the matrix so that each row sum to 1. If False, then
        the entries will be the counts of transitions.

    Returns
    -------
    transmtx : numpy.ndarray
        The transition matrix computed from the trajectory
    """
    transmtx = np.zeros([N, N])
    for i in range(1, len(traj)):
        transmtx[traj[i - 1], traj[i]] += 1   # counts of transitions
    if normalize is True:
        transmtx /= np.sum(transmtx, axis=1)[:, None]   # normalize the transition matrix
        transmtx[np.isnan(transmtx)] = 0   # for non-sampled state, there could be nan due to 0/0

    return transmtx


def plot_rep_trajs(trajs, fig_name, dt=None, stride=None):
    """
    Plots the time series of replicas visited by each configuration in a single plot.

    Parameters
    ----------
    trajs : list
        A list of arrays that represent the replica space trajectories of all configurations.
    fig_name : str
        The file name of the png file to be saved (with the extension).
    dt : float or None, optional
        One trajectory timestep in ps. If None, it assumes there are no timeframes but MC steps.
    stride : int, optional
        The stride for plotting the time series. The default is 100 if the length of
        any trajectory has more than one million frames. Otherwise, it will be 1. Typically
        plotting more than 10 million frames can take a lot of memory.
    """
    n_sim = len(trajs)
    cmap = plt.cm.ocean  # other good options are CMRmap, gnuplot, terrain, turbo, brg, etc.
    colors = [cmap(i) for i in np.arange(n_sim) / n_sim]

    if dt is None:
        x = np.arange(len(trajs[0]))
    else:
        x = np.arange(len(trajs[0])) * dt
        if max(x) >= 10000:
            x = x / 1000   # convert to ns
            units = 'ns'
        else:
            units = 'ps'

    if stride is None:
        if len(trajs[0]) > 1000000:
            stride = 100
        else:
            stride = 1

    fig = plt.figure()
    ax = fig.add_subplot(111)
    for i in range(n_sim):
        if len(trajs[0]) >= 100:  # don't show the markers
            plt.plot(x[::stride], trajs[i][::stride], color=colors[i], label=f'Configuration {i}')
        else:
            plt.plot(x[::stride], trajs[i][::stride], color=colors[i], label=f'Configuration {i}', marker='o')

    if dt is None:
        plt.xlabel('MC moves')
    else:
        plt.xlabel(f'Time ({units})')

    plt.ylabel('Replica')
    plt.grid()
    plt.legend()
    ax.yaxis.set_major_locator(MaxNLocator(integer=True))
    plt.savefig(f'{fig_name}', dpi=600)


def plot_state_trajs(trajs, state_ranges, fig_name, dt=None, stride=1):
    """
    Plots the time series of states visited by each configuration in a subplot.

    Parameters
    ----------
    trajs : list
        A list of arrays that represent the state space trajectories of all configurations.
    state_ranges : list
        A list of lists of state indices. (Like the attribute :code:`state_ranges` in :code:`EnsemblEXE`.)
    fig_name : str
        The file name of the png file to be saved (with the extension).
    dt : float or None, optional
        The time interval between consecutive frames of the trajectories. If None, it is assumed
        that the trajectories are in terms of Monte Carlo (MC) moves instead of timeframes, and
        the x-axis label is set to 'MC moves'. Default is None.
    stride : int, optional
        The stride for plotting the time series. The default is 10 if the length of
        any trajectory has more than 100,000 frames. Otherwise, it will be 1. Typically
        plotting more than 10 million frames can take a lot of memory.
    """
    n_sim = len(trajs)
    cmap = plt.cm.ocean  # other good options are CMRmap, gnuplot, terrain, turbo, brg, etc.
    colors = [cmap(i) for i in np.arange(n_sim) / n_sim]

    if dt is None:
        x = np.arange(len(trajs[0]))
    else:
        x = np.arange(len(trajs[0])) * dt
        if max(x) >= 10000:
            x = x / 1000   # convert to ns
            units = 'ns'
        else:
            units = 'ps'

    if stride is None:
        if len(trajs[0]) > 100000:
            stride = 10
        else:
            stride = 1

    # x_range = [-5, len(trajs[0]) - 1 + 5]
    x_range = [np.min(x), np.max(x)]
    y_range = [-0.2, np.max(trajs) + 0.2]
    n_configs = len(trajs)
    n_rows, n_cols = utils.get_subplot_dimension(n_configs)
    _, ax = plt.subplots(nrows=n_rows, ncols=n_cols, figsize=(2.5 * n_cols, 2.5 * n_rows))
    for i in range(n_configs):
        plt.subplot(n_rows, n_cols, i + 1)

        # First color different regions showing alchemical ranges
        for j in range(n_configs):
            bounds = [list(state_ranges[j])[0], list(state_ranges[j])[-1]]
            if j == 0:
                bounds[0] -= 0.5
            if j == n_configs - 1:
                bounds[1] += 0.5
            plt.fill_between(x_range, y1=bounds[1], y2=bounds[0], color=colors[j], alpha=0.1)

        if len(trajs[0]) > 100000:
            linewidth = 0.01
        else:
            linewidth = 1  # this is the default

        # Finally, plot the trajectories
        linewidth = 1  # this is the default
        plt.plot(x[::stride], trajs[i][::stride], color=colors[i], linewidth=linewidth)
        if dt is None:
            plt.xlabel('MC moves')
        else:
            plt.xlabel(f'Time ({units})')

        plt.ylabel('State')
        plt.title(f'Configuration {i}', fontweight='bold')
        if len(trajs[0]) >= 10000:
            plt.ticklabel_format(style='sci', axis='x', scilimits=(0, 0))
        plt.xlim(x_range)
        plt.ylim(y_range)
        plt.grid()

    # Remove redundant subplots
    n_rm = n_cols * n_rows - n_configs
    for i in range(n_rm):
        ax.flat[-1 * (i + 1)].set_visible(False)

    plt.tight_layout()
    plt.savefig(f'{fig_name}', dpi=600)


def plot_state_hist(trajs, state_ranges, fig_name):
    """
    Plots the histograms of the state index for each configuration.

    Parameters
    ----------
    trajs : list
         A list of arrays that represent the state space trajectories of all configurations.
    state_ranges : list
        A list of lists of state indices. (Like the attribute :code:`state_ranges` in :obj:`.EnsembleEXE`.)
    fig_name : str
        The file name of the png file to be saved (with the extension).
    """
    n_configs = len(trajs)
    cmap = plt.cm.ocean  # other good options are CMRmap, gnuplot, terrain, turbo, brg, etc.
    colors = [cmap(i) for i in np.arange(n_configs) / n_configs]

    fig = plt.figure()
    ax = fig.add_subplot(111)
    lower_bound = min(trajs[0]) - 0.5
    upper_bound = max(trajs[-1]) + 0.5
    for i in range(len(trajs)):
        plt.hist(trajs[i], np.arange(lower_bound, upper_bound + 1, 1), label=f'Configuration {i}', alpha=0.5, edgecolor='black', color=colors[i])  # noqa: E501
    plt.xticks(range(max(state_ranges[-1]) + 1))

    # Here we color the different regions to show alchemical ranges
    y_min, y_max = ax.get_ylim()
    for i in range(n_configs):
        bounds = [list(state_ranges[i])[0], list(state_ranges[i])[-1]]
        if i == 0:
            bounds[0] -= 0.5
        if i == n_configs - 1:
            bounds[1] += 0.5
        plt.fill_betweenx([y_min, y_max], x1=bounds[1] + 0.5, x2=bounds[0] - 0.5, color=colors[i], alpha=0.1, zorder=0)
    plt.xlim([lower_bound, upper_bound])
    plt.ylim([y_min, y_max])
    plt.xlabel('State index')
    plt.ylabel('Count')
    plt.grid()
    plt.legend()
    plt.savefig(f'{fig_name}', dpi=600)


def plot_transit_time(trajs, N, fig_prefix=None, dt=None, folder='.'):
    """
    Caclulcates and plots the average transit times for each configuration, including the time
    it takes from states 0 to k, from k to 0 and from 0 to k back to 0 (i.e. round-trip time).
    If there are more than 100 round-trips, 3 histograms corresponding to t_0k, t_k0 and t_roundtrip
    will be generated.

    Parameters
    ----------
    trajs : list
        A list of arrays that represent the state space trajectories of all configurations.
    N : int
        The total number of states in the whole alchemical range.
    fig_prefix : str
        A prefix to use for all generated figures.
    dt : float or None, optional
        One trajectory timestep in ps. If None, it assumes there are no timeframes but MC steps.
    folder : str, optional
        The directory for saving the figures

    Returns
    -------
    t_0k_list : list
        A list of transit time from states 0 to k for each configuration.
    t_k0_list : list
        A list of transit time from states k to 0 for each configuration.
    t_roundtrip_list : list
        A list of round-trip times for each configuration.
    units : str
        The units of the times.
    """
    if dt is None:
        x = np.arange(len(trajs[0]))
        units = 'step'
    else:
        x = np.arange(len(trajs[0])) * dt
        if max(x) >= 10000:
            x = x / 1000   # convert to ns
            units = 'ns'
        else:
            units = 'ps'

    # The lists below are for storing data corresponding to different configurations.
    t_0k_list, t_k0_list, t_roundtrip_list = [], [], []
    t_0k_avg, t_k0_avg, t_roundtrip_avg = [], [], []

    sci = False  # whether to use scientific notation in the y-axis in the plot
    for i in range(len(trajs)):
        traj = trajs[i]
        last_visited = None   # last visited end
        k = N - 1
        t_0, t_k = [], []   # time frames visting states 0 and k (k is the other end)

        # time spent from statkes 0 to k, k to 0 and the round-trip time (from 0 to k to 0)
        t_0k, t_k0, t_roundtrip = [], [], []

        end_0_found, end_k_found = None, None
        for t in range(len(traj)):
            if traj[t] == 0:
                end_0_found = True
                if last_visited != 0:
                    t_0.append(t)
                    if last_visited == k:
                        t_k0.append(t - t_k[-1])
                last_visited = 0
            if traj[t] == k:
                end_k_found = True
                if last_visited != k:
                    t_k.append(t)
                    if last_visited == 0:
                        t_0k.append(t - t_0[-1])
                last_visited = k

        # Here we figure out the round-trip time from t_0k and t_k0.
        if len(t_0k) != len(t_k0):   # then it must be len(t_0k) = len(t_k0) + 1 or len(t_k0) = len(t_0k) + 1, so we drop the last element of the larger list  # noqa: E501
            if len(t_0k) > len(t_k0):
                t_0k.pop()
            else:
                t_k0.pop()
        t_roundtrip = list(np.array(t_0k) + np.array(t_k0))

        if end_0_found is True and end_k_found is True:
            if dt is not None:
                units = 'ps'
                t_0k = list(np.array(t_0k) * dt)  # units: ps
                t_k0 = list(np.array(t_k0) * dt)  # units: ps
                t_roundtrip = list(np.array(t_roundtrip) * dt)  # units: ps
                if np.max([t_0k, t_k0, t_roundtrip]) >= 10000:
                    units = 'ns'
                    t_0k = list(np.array(t_0k) / 1000)   # units: ns
                    t_k0 = list(np.array(t_k0) / 1000)   # units: ns
                    t_roundtrip = list(np.array(t_roundtrip) / 1000)   # units: ns

            t_0k_list.append(t_0k)
            t_0k_avg.append(np.mean(t_0k))

            t_k0_list.append(t_k0)
            t_k0_avg.append(np.mean(t_k0))

            t_roundtrip_list.append(t_roundtrip)
            t_roundtrip_avg.append(np.mean(t_roundtrip))

            if sci is False and np.max([t_0k, t_k0, t_roundtrip]) >= 10000:
                sci = True
        else:
            t_0k_list.append([])
            t_k0_list.append([])
            t_roundtrip_list.append([])

    # Now we plot! (If there are no events, the figures will just be blank)
    meta_list = [t_0k_list, t_k0_list, t_roundtrip_list]
    y_labels = [
        f'Average transit time from states 0 to k ({units})',
        f'Average transit time from states k to 0 ({units})',
        f'Average round-trip time ({units})',
    ]
    fig_names = ['t_0k.png', 't_k0.png', 't_roundtrip.png']
    for t in range(len(meta_list)):
        t_list = meta_list[t]
        if all(not x for x in t_list):
            # If the nested list is empty, no plots will be generated.
            pass
        else:
            len_list = [len(i) for i in t_list]
            if np.max(len_list) <= 10:
                marker = 'o'
            else:
                marker = ''

            plt.figure()
            for i in range(len(t_list)):    # t_list[i] is the list for configuration i
                plt.plot(np.arange(len(t_list[i])) + 1, t_list[i], label=f'Configuration {i}', marker=marker)
            if np.array(t_list).shape != (1, 0):  # at least one configuration has at least one event
                if np.max(np.max((t_list))) >= 10000:
                    plt.ticklabel_format(style='sci', axis='y', scilimits=(0, 0))
            plt.xlabel('Event index')
            plt.ylabel(f'{y_labels[t]}')
            plt.grid()
            plt.legend()
            if fig_prefix is None:
                plt.savefig(f'{folder}/{fig_names[t]}')
            else:
                plt.savefig(f'{folder}/{fig_prefix}_{fig_names[t]}', dpi=600)

            lens = [len(t_list[i]) for i in range(len(t_list))]
            if np.min(lens) >= 100:  # plot a histogram
                counts, bins = np.histogram(t_list[i])

                plt.figure()
                for i in range(len(t_list)):
                    plt.hist(t_list[i], bins=int(len(t_list[i]) / 20), label=f'Configuration {i}')
                    if max(counts) >= 10000:
                        plt.ticklabel_format(style='sci', axis='y', scilimits=(0, 0))
                plt.xlabel(f'{y_labels[t]}')
                plt.ylabel('Event count')
                plt.grid()
                plt.legend()
                if fig_prefix is None:
                    plt.savefig(f'{folder}/hist_{fig_names[t]}', dpi=600)
                else:
                    plt.savefig(f'{folder}/{fig_prefix}_hist_{fig_names[t]}', dpi=600)

    return t_0k_list, t_k0_list, t_roundtrip_list, units



####################################################################
#                                                                  #
#    ensemble_md,                                                  #
#    a python package for running GROMACS simulation ensembles     #
#                                                                  #
#    Written by Wei-Tse Hsu <wehs7661@colorado.edu>                #
#    Copyright (c) 2022 University of Colorado Boulder             #
#                                                                  #
####################################################################
"""
The :obj:`.ensemble_EXE` module provides functions for setting up and ensemble of expanded ensemble.
"""
import os
import sys
import copy
import yaml
import shutil
import random
import numpy as np
from mpi4py import MPI
from itertools import combinations
from collections import OrderedDict
from alchemlyb.parsing.gmx import extract_dHdl
from alchemlyb.parsing.gmx import _get_headers as get_headers
from alchemlyb.parsing.gmx import _extract_dataframe as extract_dataframe

import gmxapi as gmx
import ensemble_md
from ensemble_md.utils import utils
from ensemble_md.utils import gmx_parser
from ensemble_md.utils.exceptions import ParameterError


rank = MPI.COMM_WORLD.Get_rank()  # Note that this is a GLOBAL variable


class EnsembleEXE:
    """
    This class provides a variety of functions useful for setting up and running
    an ensemble of expanded ensemble. Upon instantiation, all parameters in the YAML
    file will be assigned to an attribute in the class. In addition to these variables,
    below is a list of attributes of the class. (All the the attributes are assigned by
    :obj:`set_params` unless otherwise noted.)

    :ivar yaml: The input YAML file used to instantiate the class. Assigned by the :code:`__init__` function.
    :ivar warnings: Warnings about parameter specification in either YAML or MDP files.
    :ivar reformatted_mdp: Whether the templated MDP file has been reformatted by replacing hyphens
        with underscores or not.
    :ivar template: The instance of the :obj:`MDP` class based on the template MDP file.
    :ivar nsteps: The number of steps per iteration.
    :ivar dt: The simulation timestep in ps.
    :ivar temp: The simulation temperature in Kelvin.
    :ivar fixed_weights: Whether the weights will be fixed during the simulation (according to the template MDP file).
    :ivar kT: 1 kT in kJ/mol at the simulation temperature.
    :ivar lambda_types: The types of lambda variables involved in expanded ensemble simulations, e.g.
        :code:`fep_lambdas`, :code:`mass_lambdas`, :code:`coul_lambdas`, etc.
    :ivar n_tot: The total number of states for all replicas.
    :ivar n_sub: The numbmer of states for each replica. The current implementation assumes homogenous replicas.
    :ivar state_ranges: A list of list of state indices for each replica.
    :ivar equil: A list of times it took to equilibrated the weights for different replicas.
    :ivar lambda_dict: A dictionary with keys being the tuples of coupling parameters used in each replicas and
        values being the corresponding global index (starting from 0). Assigned by :obj:`map_lambda2state`.
    :ivar lambda_ranges: A list of lambda vectors of state range of each replica. Assigned by :obj:`map_lambda2state`.
    :ivar n_rejected: The number of proposed exchanges that have been rejected. Updated by :obj:`accept_or_reject`.
    :ivar n_swap_attempts: The number of swaps attempted so far. This does not include the cases
        where there is no swappable pair. Updated by :obj:`get_swapping_pattern`.
    :ivar rep_trajs: The replica-space trajectories of all replicas.
    :ivar configs: A list that thows the current configuration index that each replica is sampling.
    :ivar g_vecs: The time series of the (processed) whole-range alchemical weights. If no weight combination is
        applied, this list will just be a list of :code:`None`'s.
    :ivar get_u_nk: Whether to get the :math:`u_{nk}` dataset from the DHDL files. Only meaningful during
        data analysis and if :code:`df_method` is specified.
    :ivar get_dHdl: Whether to get the :math:`dH/dλ` dataset from the DHDL files. Only meaningful
        during data analysis and if :code:`df_method` is specified.
    """

    def __init__(self, yaml_file):
        self.yaml = yaml_file
        self.set_params()

    def set_params(self):
        """
        Sets up or reads in the user-defined parameters from a yaml file and an MDP template.
        This function is called to instantiate the class in the :code:`__init__` function of
        class. Specifically, it does the following:

          1. Sets up constants.
          2. Reads in parameters from a YAML file.
          3. Handles YAML parameters.
          4. Checks if the parameters in the YAML file are well-defined.
          5. Reformats the input MDP file to replace all hyphens with underscores.
          6. Reads in parameters from the MDP template.

        After instantiation, the class instance will have attributes corresponding to
        each of the parameters specified in the YAML file. For a full list of the parameters that
        can be specified in the YAML file, please refer to :ref:`doc_parameters`.

        :param yaml_file: The file name of the YAML file for specifying the parameters for EEXE.
        :type yaml_file: str

        :raises ParameterError:

              - If a required parameter is not specified in the YAML file.
              - If a specified parameter is not recognizable.
              - If a specified weight combining scheme is not available.
              - If a specified acceptance scheme is not available.
              - If a specified free energy estimator is not available.
              - If a specified method for error estimation is not available.
              - If an integer parameter is not an integer.
              - If a positive parameter is not positive.
              - If a non-negative parameter is negative.
              - If an invalid MDP file is detected.
        """
        self.warnings = []  # Store warnings, if any.

        # Step 0: Set up constants
        k = 1.380649e-23
        NA = 6.0221408e23

        # Step 1: Read in parameters from the YAML file.
        with open(self.yaml) as f:
            params = yaml.load(f, Loader=yaml.FullLoader)

        for attr in params:
            setattr(self, attr, params[attr])

        # Step 2: Handle the YAML parameters
        required_args = [
            "gro",
            "top",
            "mdp",
            "parallel",
            "n_sim",
            "n_iter",
            "s",
        ]
        for i in required_args:
            if hasattr(self, i) is False or getattr(self, i) is None:
                raise ParameterError(
                    f"Required parameter '{i}' not specified in {self.yaml}."
                )  # noqa: F405

        # Key: Optional argument; Value: Default value
        optional_args = {
            "nst_sim": None,
            "proposal": 'exhaustive',
            "acceptance": "metropolis",
            "w_scheme": None,
            "N_cutoff": 1000,
            "n_ex": 'N^3',   # only active for multiple swaps.
            "verbose": True,
            "runtime_args": None,
            "n_ckpt": 100,
            "msm": False,
            "free_energy": False,
            "df_spacing": 1,
            "df_method": "MBAR",
            "err_method": "propagate",
            "n_bootstrap": 50,
            "seed": None,
        }
        for i in optional_args:
            if hasattr(self, i) is False or getattr(self, i) is None:
                setattr(self, i, optional_args[i])

        # all_args: Arguments that can be specified in the YAML file.
        all_args = required_args + list(optional_args.keys())
        for i in params:
            if i not in all_args:
                self.warnings.append(f'Warning: Parameter "{i}" specified in the input YAML file is not recognizable.')

        # Step 3: Check if the parameters in the YAML file are well-defined
        if self.proposal not in [None, 'single', 'neighboring', 'exhaustive', 'multiple']:
            raise ParameterError("The specified proposal scheme is not available. Available options include 'single', 'neighboring', 'exhaustive', and 'multiple'.")  # noqa: E501

        if self.acceptance not in [None, 'same-state', 'same_state', 'metropolis', 'metropolis-eq', 'metropolis_eq']:
            raise ParameterError("The specified acceptance scheme is not available. Available options include 'same-state', 'metropolis', and 'metropolis-eq'.")  # noqa: E501

        if self.w_scheme not in [None, 'mean', 'geo-mean', 'g-diff']:
            raise ParameterError("The specified weight combining scheme is not available. Available options include None, 'mean', 'geo-mean'/'geo_mean' and 'g-diff/g_diff'.")  # noqa: E501

        if self.df_method not in [None, 'TI', 'BAR', 'MBAR']:
            raise ParameterError("The specified free energy estimator is not available. Available options include 'TI', 'BAR', and 'MBAR'.")  # noqa: E501

        if self.err_method not in [None, 'propagate', 'bootstrap']:
            raise ParameterError("The specified method for error estimation is not available. Available options include 'propagate', and 'bootstrap'.")  # noqa: E501

        params_int = ['n_sim', 'n_iter', 's', 'N_cutoff', 'df_spacing', 'n_ckpt', 'n_bootstrap']  # integer parameters  # noqa: E501
        if self.nst_sim is not None:
            params_int.append('nst_sim')
        if self.n_ex != 'N^3':  # no need to add "and self.proposal == 'multiple' since if multiple swaps are not used, n_ex=1"  # noqa: E501
            params_int.append('n_ex')
        if self.seed is not None:
            params_int.append('seed')
        for i in params_int:
            if type(getattr(self, i)) != int:
                raise ParameterError(f"The parameter '{i}' should be an integer.")

        params_pos = ['n_sim', 'n_iter', 's', 'n_ckpt', 'df_spacing', 'n_bootstrap']  # positive parameters
        if self.nst_sim is not None:
            params_pos.append('nst_sim')
        for i in params_pos:
            if getattr(self, i) <= 0:
                raise ParameterError(f"The parameter '{i}' should be positive.")

        if self.n_ex != 'N^3' and self.n_ex < 0:
            raise ParameterError("The parameter 'n_ex' should be non-negative.")

        if self.N_cutoff < 0 and self.N_cutoff != -1:
            raise ParameterError("The parameter 'N_cutoff' should be non-negative unless no histogram correction is needed, i.e. N_cutoff = -1.")  # noqa: E501

        params_str = ['gro', 'top', 'mdp']
        for i in params_str:
            if type(getattr(self, i)) != str:
                raise ParameterError(f"The parameter '{i}' should be a string.")

        params_bool = ['parallel', 'verbose', 'msm']
        for i in params_bool:
            if type(getattr(self, i)) != bool:
                raise ParameterError(f"The parameter '{i}' should be a boolean variable.")

        # Step 4: Reformat the input MDP file to replace all hypens with underscores.
        self.reformat_MDP()

        # Step 5: Read in parameters from the MDP template
        self.template = gmx_parser.MDP(self.mdp)
        self.nsteps = self.template["nsteps"]  # will be overwritten by self.nst_sim if nst_sim is specified.
        self.dt = self.template["dt"]  # ps
        self.temp = self.template["ref_t"]

        if self.nst_sim is None:
            self.nst_sim = self.nsteps

        if 'wl_scale' in self.template.keys():
            if isinstance(self.template['wl_scale'], np.ndarray):
                # If wl_scale in the MDP file is a blank (i.e. fixed weights), mdp['wl_scale'] will be an empty array.
                # This is the only case where mdp['wl_scale'] is a numpy array.
                self.fixed_weights = True
            else:
                self.fixed_weights = False
        else:
            self.fixed_weights = True

        if 'lmc_seed' in self.template and self.template['lmc_seed'] != -1:
            self.warnings.append('Warning: We recommend setting lmc_seed as -1 so the random seed is different for each iteration.')  # noqa: E501

        if 'gen_seed' in self.template and self.template['gen_seed'] != -1:
            self.warnings.append('Warning: We recommend setting gen_seed as -1 so the random seed is different for each iteration.')  # noqa: E501

        if 'symmetrized_transition_matrix' in self.template and self.template['symmetrized_transition_matrix'] == 'yes':  # noqa: E501
            self.warnings.append('Warning: We recommend setting symmetrized-transition-matrix to no instead of yes.')

        if self.template['nstlog'] > self.nst_sim:
            raise ParameterError(
                'The parameter "nstlog" should be equal to or smaller than "nst_sim" specified in the YAML file so that the sampling information can be parsed.')  # noqa: E501

        # Step 6: Set up derived parameters
        # 6-1. kT in kJ/mol
        self.kT = k * NA * self.temp / 1000  # 1 kT in kJ/mol

        # 6-2. Figure out what types of lambda variables are involved
        # Here is we possible lambda types in the order read by GROMACS, which is likely also the order when being printed to the log file.  # noqa: E501
        # See https://gitlab.com/gromacs/gromacs/-/blob/main/src/gromacs/gmxpreprocess/readir.cpp#L2543
        lambdas_types_all = ['fep_lambdas', 'mass_lambdas', 'coul_lambdas', 'vdw_lambdas', 'bonded_lambdas', 'restraint_lambdas', 'temperature_lambdas']  # noqa: E501
        self.lambda_types = []  # lambdas specified in the MDP file
        for i in lambdas_types_all:
            if i in self.template.keys():  # there shouldn't be parameters like "fep-lambdas" after reformatting the MDP file  # noqa: E501
                self.lambda_types.append(i)

        # 6-3. Total # of states: n_tot = n_sub * n_sim - (n_overlap) * (n_sim - 1), where n_overlap = n_sub - s
        self.n_tot = len(self.template[self.lambda_types[0]])

        # 6-4. Number of states of each replica (assuming the same for all rep)
        self.n_sub = self.n_tot - self.s * (self.n_sim - 1)
        if self.n_sub < 1:
            raise ParameterError(
                f"There must be at least two states for each replica (current value: {self.n_sub}). The current specified configuration (n_tot={self.n_tot}, n_sim={self.n_sim}, s={self.s}) does not work for EEXE.")  # noqa: E501

        # 6-5. A list of sets of state indices
        start_idx = [i * self.s for i in range(self.n_sim)]
        self.state_ranges = [list(np.arange(i, i + self.n_sub)) for i in start_idx]

        # 6-6. A list of time it took to get the weights equilibrated
        self.equil = [-1 for i in range(self.n_sim)]   # -1 means unequilibrated

        # 6-7. Map the lamda vectors to state indices
        self.map_lambda2state()

        # 6-8. For counting the number swap attempts and the rejected ones
        self.n_rejected = 0
        self.n_swap_attempts = 0

        # 6-9. Replica space trajectories. For example, rep_trajs[0] = [0, 2, 3, 0, 1, ...] means
        # that configuration 0 transitioned to replica 2, then 3, 0, 1, in iterations 1, 2, 3, 4, ...,
        # respectively. The first element of rep_traj[i] should always be i.
        self.rep_trajs = [[i] for i in range(self.n_sim)]

        # 6-10. configs shows the current configuration that each replica is sampling.
        # For example, self.configs = [0, 2, 1, 3] means that configurations 0, 2, 1, and 3 are
        # in replicas, 0, 1, 2, 3, respectively. This list will be constantly updated during the simulation.
        self.configs = list(range(self.n_sim))

        # 6-11. The time series of the (processed) whole-range alchemical weights
        # If no weight combination is applied, self.g_vecs will just be a list of None's.
        self.g_vecs = []

        # 6-12. Data analysis
        if self.df_method == 'MBAR':
            self.get_u_nk = True
            self.get_dHdl = False
        else:
            self.get_u_nk = False
            self.get_dHdl = True

    def print_params(self, params_analysis=False):
        """
        Prints important parameters related to the EXEE simulation.

        Parameters
        ----------
        params_analysis : bool, optional
            If True, additional parameters related to data analysis will be printed. Default is False.
        """
        print("Important parameters of EXEE")
        print("============================")
        print(f"Python version: {sys.version}")
        print(f"gmxapi version: {gmx.__version__}")
        print(f"ensemble_md version: {ensemble_md.__version__}")
        print(f'Simulation inputs: {self.gro}, {self.top}, {self.mdp}')
        print(f"Verbose log file: {self.verbose}")
        print(f"Whether the replicas run in parallel: {self.parallel}")
        print(f"Proposal scheme: {self.proposal}")
        print(f"Acceptance scheme for swapping simulations: {self.acceptance}")
        print(f"Scheme for combining weights: {self.w_scheme}")
        print(f"Histogram cutoff: {self.N_cutoff}")
        print(f"Number of replicas: {self.n_sim}")
        print(f"Number of iterations: {self.n_iter}")
        print(f"Number of attempted swaps in one exchange interval: {self.n_ex}")
        print(f"Length of each replica: {self.dt * self.nst_sim} ps")
        print(f"Frequency for checkpointing: {self.n_ckpt} iterations")
        print(f"Total number of states: {self.n_tot}")
        print(f"Additional runtime arguments: {self.runtime_args}")
        print("Alchemical ranges of each replica in EEXE:")
        for i in range(self.n_sim):
            print(f"  - Replica {i}: States {self.state_ranges[i]}")

        if params_analysis is True:
            print()
            print(f"Whether to build Markov state models and perform relevant analysis: {self.msm}")
            print(f"Whether to perform free energy calculations: {self.free_energy}")
            print(f"The step to used in subsampling the DHDL data in free energy calculations, if any: {self.df_spacing}")  # noqa: E501
            print(f"The chosen free energy estimator for free energy calculations, if any: {self.df_method}")
            print(f"The method for estimating the uncertainty of free energies in free energy calculations, if any: {self.err_method}")  # noqa: E501
            print(f"The number of bootstrap iterations in the boostrapping method, if used: {self.n_bootstrap}")
            print(f"The random seed to use in bootstrapping, if used: {self.seed}")

        if self.reformatted_mdp is True:
            print("Note that the input MDP file has been reformatted by replacing hypens with underscores. The original mdp file has been renamed as *backup.mdp.")  # noqa: E501

    def reformat_MDP(self):
        """
        Reformats an MDP file so that all hyphens in the parameter names are replaced by underscores.
        If the input MDP file contains hyphens in its parameter names, the class attribue :code:`self.reformatted`
        will be set to :code:`True`. In this case, the new MDP object with reformatted parameter names will be
        written to the original file paththe file, while the original file will be renamed with a
        :code:`_backup` suffix. If the input MDP file contains underscores in its parameter names, the function sets
        the class attribute :code:`self.reformatted_mdp` to :code:`False`.
        """
        params = gmx_parser.MDP(self.mdp)

        odict = OrderedDict([(k.replace('-', '_'), v) for k, v in params.items()])
        params_new = gmx_parser.MDP(None, **odict)

        if params_new.keys() == params.keys():
            self.reformatted_mdp = False  # no need to reformat the file
        else:
            self.reformatted_mdp = True
            new_name = self.mdp.split('.mdp')[0] + '_backup.mdp'
            shutil.move(self.mdp, new_name)
            params_new.write(self.mdp)

    def map_lambda2state(self):
        """
        Define attrbitues that map vectors of coupling parameters to state indices.

        Attributes
        ----------
        lambda_dict : dict
            A dictionary whose keys are tuples of coupling parameters and
            values are the corresponding GLOBAL state indices (starting from 0).
        lambda_ranges : list
            A list of lambda vectors of the state range of each replica.
        """
        self.lambda_dict = {}  # key: vector of coupling parameters, value: state index
        for i in range(self.n_tot):
            key = tuple([self.template[j][i] for j in self.lambda_types])
            self.lambda_dict[key] = i
        self.lambda_ranges = [[list(self.lambda_dict.keys())[j] for j in self.state_ranges[i]]for i in range(len(self.state_ranges))]  # noqa: E501

    def initialize_MDP(self, idx):
        """
        Initializes the MDP object for generating MDP files for a replica based on the MDP template.
        This function should be called only for generating MDP files for the FIRST iteration
        and it has nothing to do with whether the weights are fixed or equilibrating.
        It is assumed that the MDP template has all the common parameters of all replicas.

        Parameters
        ----------
        idx : int
            Index of the simulation whose MDP parameters need to be initialized.

        Returns
        -------
        MDP : :obj:`.gmx_parser.MDP` obj
            An updated object of :obj:`.gmx_parser.MDP` that can be used to write MDP files.
        """
        MDP = copy.deepcopy(self.template)
        MDP["nsteps"] = self.nst_sim

        start_idx = idx * self.s
        for i in self.lambda_types:
            MDP[i] = self.template[i][start_idx:start_idx + self.n_sub]

        if "init_lambda_weights" in self.template:
            MDP["init_lambda_weights"] = self.template["init_lambda_weights"][start_idx:start_idx + self.n_sub]

        return MDP

    def update_MDP(self, new_template, sim_idx, iter_idx, states, wl_delta, weights):
        """
        Updates the MDP file for a new iteration based on the new MDP template coming from the previous iteration.
        Note that if the weights got equilibrated in the previous iteration, then the weights will be fixed
        at these equilibrated values for all the following iterations.

        Parameters
        ----------
        new_template : str
            The new MDP template file. Typically the MDP file of the previous iteration.
        sim_idx : int
            The index of the simulation whose MDP parameters need to be updated.
        iter_idx : int
            The index of the iteration to be performed later.
        states : list
            A list of last sampled states of all simulaitons in the previous iteration.
        wl_delta : list
            A list of final Wang-Landau incrementors of all simulations.
        weights : list
            A list of lists final weights of all simulations.

        Return
        ------
        MDP : :obj:`.gmx_parser.MDP` obj
            An updated object of :obj:`.gmx_parser.MDP` that can be used to write MDP files.
        """
        new_template = gmx_parser.MDP(new_template)  # turn into a gmx_parser.MDP object
        MDP = copy.deepcopy(new_template)
        MDP["tinit"] = self.nst_sim * self.dt * iter_idx
        MDP["nsteps"] = self.nst_sim
        MDP["init_lambda_state"] = (states[sim_idx] - sim_idx * self.s)  # 2nd term is for shifting from the global to local index.  # noqa: E501
        MDP["init_lambda_weights"] = weights[sim_idx]
        MDP["init_wl_delta"] = wl_delta[sim_idx]

        if self.equil[sim_idx] == -1:   # the weights haven't been equilibrated
            MDP["init_wl_delta"] = wl_delta[sim_idx]
        else:
            MDP["lmc_stats"] = "no"
            MDP["wl_scale"] = ""
            MDP["wl_ratio"] = ""
            MDP["init_wl_delta"] = ""
            MDP["lmc_weights_equil"] = ""
            MDP["weight_equil_wl_delta"] = ""

        return MDP

    def extract_final_dhdl_info(self, dhdl_files):
        """
        For all the replica simulations, finds the last sampled state
        and the corresponding lambda values from a dhdl file.

        Parameters
        ----------
        dhdl_files : list
            A list of GROMACS DHDL file names.

        Returns
        -------
        states : list
            A list of the global indices of the last sampled states of all simulaitons.
        """
        states, lambda_vecs = [], []
        print("\nBelow are the final states being visited:")
        for j in range(self.n_sim):
            dhdl = extract_dHdl(dhdl_files[j], T=self.temp)
            lambda_vecs.append(dhdl.index[-1][1:])
            states.append(self.lambda_dict[lambda_vecs[-1]])  # absolute order
            print(
                f"  Simulation {j}: Global state {states[j]}, (coul, vdw) = \
                {list(self.lambda_dict.keys())[list(self.lambda_dict.values()).index(states[j])]}"
            )
        print('\n', end='')

        return states

    def extract_final_log_info(self, log_files):
        """
        Extracts the following information for each replica simulation from a given list of GROMACS LOG files:

          - The final Wang-Landau incrementors.
          - The final lists of weights.
          - The final lists of counts.
          - Whether the weights were equilibrated in the simulations.

        Parameters
        ----------
        log_files : list
            A list of file paths to GROMACS LOG files.

        Returns
        -------
        wl_delta : list
            A list of final Wang-Landau incrementors of all simulations.
        weights : list
            A list of lists of final weights of all simulations.
        counts : list
            A list of lists of final counts of all simulations.
        """
        wl_delta, weights, counts = [], [], []

        # 2. Find the final Wang-Landau incrementors and weights
        for j in range(self.n_sim):
            if self.verbose:
                print(f'Parsing {log_files[j]} ...')
            result = gmx_parser.parse_log(log_files[j])
            wl_delta.append(result[0])
            weights.append(result[1])
            counts.append(result[2])

            # In Case 3 described in the docstring of parse_log (fixed-weights),
            # result[3] will be 0 but it will never be passed to self.equil[j]
            # because once self.equil[j] is not -1, we stop updating. This way, we can keep
            # the time when the weights get equilibrated all the way.
            if self.equil[j] == -1:
                # At this point self.equil is the equilibration status BEFORE the last iteration
                # while result[3] is the equilibration status AFTER finishing the last iteraiton.
                # For any replicas where weights are still equilibrating (i.e. self.equil[j] == -1)
                # we update its equilibration status.
                self.equil[j] = result[3]

        return wl_delta, weights, counts

    def identify_swappable_pairs(self, states, state_ranges):
        """
        Identify swappable pairs. By definition, a pair of simulation is considered swappable only if
        their last sampled states are in the alchemical ranges of both simulations. This is required
        to ensure that the values of involved ΔH and Δg can always be looked up from the DHDL and LOG files.
        This also automatically guarantee that the simulations to be swapped have overlapping alchemical ranges.

        Parameters
        ----------
        states : list
            A list of the global indices of the last sampled states of all simulations. This list can be
            generated by the :obj:`.extract_final_dhdl_info` method. Notably, the input list should not be
            a list that has been updated/modified by :obj:`get_swapping_pattern`, or the result will be incorrect.
        state_ranges : list of lists
            A list of state indies for all replicas. The input list can be a list updated by
            :code:`get_swapping_pattern`, especially in the case where there is a need to re-identify the
            swappable pairs after an attempted swap is accepted.

        Returns
        -------
        swappables : list
            A list of tuples representing the simulations that can be swapped.
        """
        sim_idx = list(range(self.n_sim))
        all_pairs = list(combinations(sim_idx, 2))

        # First, we identify pairs of replicas with overlapping ranges
        swappables = [i for i in all_pairs if set(state_ranges[i[0]]).intersection(set(state_ranges[i[1]])) != set()]  # noqa: E501

        # Then, from these pairs, we exclude the ones whose the last sampled states are not present in both alchemical ranges  # noqa: E501
        # In this case, U^i_n, U_^j_m, g^i_n, and g_^j_m are unknown and the acceptance cannot be calculated.
        swappables = [i for i in swappables if states[i[0]] in state_ranges[i[1]] and states[i[1]] in state_ranges[i[0]]]  # noqa: E501

        if self.proposal == 'neighboring':
            print('Note: One neighboring swap will be proposed.')
            swappables = [i for i in swappables if np.abs(i[0] - i[1]) == 1]

        return swappables

    def propose_swap(self, swappables):
        """
        Proposes a swap of coordinates between replicas by drawing samples from the swappable pairs.

        Parameters
        ----------
        swappables : list
            A list of tuples representing the simulations that can be swapped.

        Return
        ------
        swap : tuple or an empty list
            A tuple of simulation indices to be swapped. If there is no swappable pair,
            an empty list will be returned.
        """
        try:
            swap = random.choices(swappables, k=1)[0]
        except IndexError:  # no swappable pairs
            swap = []

        return swap

    def get_swapping_pattern(self, dhdl_files, states, weights):
        """
        Generates a list (:code:`swap_pattern`) that represents how the configurations should be swapped in the
        next iteration. The indices of the output list correspond to the simulation/replica indices, and the
        values represent the configuration indices in the corresponding simulation/replica. For example, if the
        swapping pattern is :code:`[0, 2, 1, 3]`, it means that in the next iteration, replicas 0, 1, 2, 3 should
        sample configurations 0, 2, 1, 3, respectively, where configurations 0, 1, 2, 3 here are defined as whatever
        configurations are in replicas 0, 1, 2, 3 in the CURRENT iteration (not iteration 0), respectively.

        Notably, when this function is called (e.g. once every iteration in an EEXE simulation), the output
        list :code:`swap_pattern` is always initialized as :code:`[0, 1, 2, 3, ...]` and gets updated once every
        attempted swap. This is different from the attribute :code:`configs`, which is only initialized at the
        very beginning of the entire EEXE simulation (iteration 0), though :code:`configs` also gets updated with
        :code:`swap_pattern` once every attempted swap in this function.

        Parameters
        ----------
        dhdl_files : list
            A list of DHDL files. The indicies in the DHDL filenames shouuld be in an ascending order, e.g.
            :code:`[dhdl_0.xvg, dhdl_1.xvg, ..., dhdl_N.xvg]`.
        states : list
            A list of last sampled states (in global indices) of ALL simulaitons. :code:`states[i]=j` means that
            the configuration in replica :code:`i` is at state :code:`j` at the time when the exchange is performed.
            This list can be generated :obj:`.extract_final_dhdl_info`.
        weights : list
            A list of lists of final weights of ALL simulations. :code:`weights[i]` corresponds to the list of weights
            used in replica :code:`i`. The list can be generated by :obj:`.extract_final_log_info`.

        Returns
        -------
        swap_pattern : list
            A list that represents how the replicas should be swapped.
        """
        if self.proposal != 'multiple':
            if self.proposal == 'exhaustive':
                n_ex = int(np.floor(self.n_sim / 2))  # This is the maximum, not necessarily the number that will always be reached.  # noqa
                n_ex_exhaustive = 0    # The actual number of swaps atttempted.
            else:
                n_ex = 1  # single swap or neighboring swap
        else:
            # multiple swaps
            if self.n_ex == 'N^3':
                n_ex = self.n_tot ** 3
            else:
                n_ex = self.n_ex

        shifts = list(self.s * np.arange(self.n_sim))
        swap_pattern = list(range(self.n_sim))   # Can be regarded as the indices of DHDL files/configurations
        state_ranges = copy.deepcopy(self.state_ranges)
        states_copy = copy.deepcopy(states)  # only for re-identifying swappable pairs given updated state_ranges
        swappables = self.identify_swappable_pairs(states, state_ranges)

        # Note that if there is only 1 swappable pair, then it will still be the only swappable pair
        # after an attempted swap is accepted. Therefore, there is no need to perform multiple swaps or re-identify
        # the new set of swappable pairs. In this case, we set n_ex back to 1.
        if len(swappables) == 1:
            if n_ex > 1:
                print('n_ex is set back to 1 since there is only 1 swappable pair.')
                n_ex = 1

        print(f"Swappable pairs: {swappables}")
        for i in range(n_ex):
            # Update the list of swappable pairs starting from the 2nd attempted swap for the exhaustive swap method.
            if self.proposal == 'exhaustive' and i >= 1:
                # Note that this should be done regardless of the acceptance/rejection of the previously drawn pairs.
                # Also note that at this point, swap is still the last attempted swap.
                swappables = [i for i in swappables if set(i).intersection(set(swap)) == set()]  # noqa: F821
                print(f'\nRemaining swappable pairs: {swappables}')

            if len(swappables) == 0 and self.proposal == 'exhaustive':
                # This should only happen when the method of exhaustive swaps is used.
                print(f'{n_ex_exhaustive} swap(s) have been attempted to exhaustively explore all possible swaps.')
                if i == 0:
                    self.n_swap_attempts += 1
                    self.n_rejected += 1
                break
            else:
                self.n_swap_attempts += 1
                if self.proposal == 'exhaustive':
                    n_ex_exhaustive += 1

                swap = self.propose_swap(swappables)
                print(f'\nProposed swap: {swap}')
                if swap == []:
                    print('No swap is proposed because there is no swappable pair at all.')
                    break  # no need to re-identify swappable pairs and draw new samples
                else:
                    if self.verbose is True and self.proposal != 'exhaustive':
                        print(f'A swap ({i + 1}/{n_ex}) is proposed between the configurations of Simulation {swap[0]} (state {states[swap[0]]}) and Simulation {swap[1]} (state {states[swap[1]]}) ...')  # noqa: E501

                    # Calculate the acceptance ratio and decide whether to accept the swap.
                    prob_acc = self.calc_prob_acc(swap, dhdl_files, states, shifts, weights)
                    swap_bool = self.accept_or_reject(prob_acc)

                    # Theoretically, in an EEXE simulation, we could either choose to swap configurations (via
                    # swapping GRO files) or replicas (via swapping MDP files). In ensemble_md package, we chose the
                    # former when implementing the EEXE algorithm. Specifically, in the CLI `run_EEXE`, `swap_pattern`
                    # is used to swap the GRO files. Therefore, when an attempted swap is accetped and `swap_pattern`
                    # is updated, we also need to update the variables `shifts`, `weights`, `dhdl_files`,
                    # `state_ranges`, `self.configs` but not anything else. Otherwise, incorrect results will be
                    # produced. To better understand this, one can refer to our unit test for get_swapping_pattern
                    # and calc_prob_acc, set checkpoints and examine why the variables should/should not be updated.

                    if swap_bool is True:
                        # The assignments need to be done at the same time in just one line.
                        # states[swap[0]], states[swap[1]] = states[swap[1]], states[swap[0]]
                        shifts[swap[0]], shifts[swap[1]] = shifts[swap[1]], shifts[swap[0]]
                        weights[swap[0]], weights[swap[1]] = weights[swap[1]], weights[swap[0]]
                        dhdl_files[swap[0]], dhdl_files[swap[1]] = dhdl_files[swap[1]], dhdl_files[swap[0]]
                        swap_pattern[swap[0]], swap_pattern[swap[1]] = swap_pattern[swap[1]], swap_pattern[swap[0]]
                        state_ranges[swap[0]], state_ranges[swap[1]] = state_ranges[swap[1]], state_ranges[swap[0]]
                        self.configs[swap[0]], self.configs[swap[1]] = self.configs[swap[1]], self.configs[swap[0]]

                        if n_ex > 1 and self.proposal == 'multiple':  # must be multiple swaps
                            # After state_ranges have been updated, we re-identify the swappable pairs.
                            # Notably, states_copy (instead of states) should be used. (They could be different.)
                            swappables = self.identify_swappable_pairs(states_copy, state_ranges)
                            print(f"  New swappable pairs: {swappables}")
                    else:
                        # In this case, there is no need to update the swappables
                        pass

                print(f'  Current list of configurations: {self.configs}')

        if self.verbose is False:
            print(f'\n{n_ex} swap(s) have been proposed.')
        print(f'\nThe finally adopted swap pattern: {swap_pattern}')
        print(f'The list of configurations sampled in each replica in the next iteration: {self.configs}')

        # Update the replica-space trajectories
        for i in range(self.n_sim):
            self.rep_trajs[i].append(self.configs.index(i))

        return swap_pattern

    def calc_prob_acc(self, swap, dhdl_files, states, shifts, weights):
        """
        Calculates the acceptance ratio given the Monte Carlo scheme for swapping the simulations.

        Parameters
        ----------
        swap : tuple
            A tuple of indices corresponding to the simulations to be swapped.
        dhdl_files : list
            A list of DHDL files, e.g. :code:`dhdl_files = ['dhdl_2.xvg', 'dhdl_1.xvg', 'dhdl_0.xvg', 'dhdl_3.xvg']`
            means that configurations 2, 1, 0, and 3 are now in replicas 0, 1, 2, 3. This can happen in multiple swaps
            when a previous swap between configurations 0 and 2 has just been accepted. Otherwise, the list of
            filenames should always be in the ascending order, e.g. :code:`['dhdl_0.xvg', 'dhdl_1.xvg', 'dhdl_2.xvg',
            dhdl_3.xvg]`.
        states : list
            A list of last sampled states (in global indices) in the DHDL files corresponding to configurations 0, 1,
            2, ... (e.g. :code:`dhdl_0.xvg`, :code:`dhdl_1.xvg`, :code:`dhdl_2.xvg`, ...)
            This list can be generated by :obj:`.extract_final_dhdl_info`.
        shifts : list
            A list of state shifts for converting global state indices to the local ones. Specifically, :code:`states`
            substracted by :code:`shifts` should be the local state indices of the last sampled states
            in :code:`dhdl_files[0]`, :code:`dhdl_files[1]`, ... (which importantly, are not necessarily
            :code:`dhdl_0.xvg`, :code:`dhdl_1.xvg`, ...). If multiple swaps are not used, then
            this should always be :code:`list(EEXE.s * np.arange(EEXE.n_sim))`.
        weights : list
            A list of lists of final weights of ALL simulations. Typiecally generated by
            :obj:`.extract_final_log_info`. :code:`weights[i]` and :code:`dhdl_files[i]` should correspond to
            the same configuration.

        Returns
        -------
        prob_acc : float
            The acceptance ratio.
        """
        if self.acceptance == "same_state" or self.acceptance == "same-state":
            if states[swap[0]] == states[swap[1]]:  # same state, swap!
                prob_acc = 1  # This must lead to an output of swap_bool = True from the function accept_or_reject
            else:
                prob_acc = 0  # This must lead to an output of swap_bool = False from the function accept_or_reject

        else:  # i.e. metropolis-eq or metropolis, which both require the calculation of dU
            # Now we calculate dU
            if self.verbose is True:
                print("  Proposing a move from (x^i_m, x^j_n) to (x^i_n, x^j_m) ...")
            f0, f1 = dhdl_files[swap[0]], dhdl_files[swap[1]]
            h0, h1 = get_headers(f0), get_headers(f1)
            data_0, data_1 = (
                extract_dataframe(f0, headers=h0).iloc[-1],
                extract_dataframe(f1, headers=h1).iloc[-1],
            )

            # \Delta H to all states at the last time frame
            # Notably, the can be regarded as H for each state since the reference state will have a value of 0 anyway.
            dhdl_0 = data_0[-self.n_sub:]
            dhdl_1 = data_1[-self.n_sub:]

            # Old local index, which will only be used in "metropolis"
            old_state_0 = states[swap[0]] - shifts[swap[0]]
            old_state_1 = states[swap[1]] - shifts[swap[1]]

            # New local index. Note that states are global indices, so we shift them back to the local ones
            new_state_0 = states[swap[1]] - shifts[swap[0]]  # new state index (local index in simulation swap[0])
            new_state_1 = states[swap[0]] - shifts[swap[1]]  # new state index (local index in simulation swap[1])

            dU_0 = (dhdl_0[new_state_0] - dhdl_0[old_state_0]) / self.kT  # U^{i}_{n} - U^{i}_{m}, i.e. \Delta U (kT) to the new state  # noqa: E501
            dU_1 = (dhdl_1[new_state_1] - dhdl_1[old_state_1]) / self.kT  # U^{j}_{m} - U^{j}_{n}, i.e. \Delta U (kT) to the new state  # noqa: E501
            dU = dU_0 + dU_1
            if self.verbose is True:
                print(
                    f"  U^i_n - U^i_m = {dU_0:.2f} kT, U^j_m - U^j_n = {dU_1:.2f} kT, Total dU: {dU:.2f} kT"
                )

            if self.acceptance == "metropolis_eq" or self.acceptance == "metropolis-eq":
                prob_acc = min(1, np.exp(-dU))
            else:  # must be 'metropolis', which consider lambda weights as well
                g0 = weights[swap[0]]
                g1 = weights[swap[1]]
                dg_0 = g0[new_state_0] - g0[old_state_0]  # g^{i}_{n} - g^{i}_{m}
                dg_1 = g1[new_state_1] - g1[old_state_1]  # g^{j}_{m} - g^{j}_{n}
                dg = dg_0 + dg_1  # kT

                # Note that simulations with different lambda ranges would have different references
                # so g^{i}_{n} - g^{j}_{n} or g^{j}_{m} - g^{i}_{m} wouldn't not make sense.
                # We therefore print g^i_n - g^i_m and g^j_m - g^j_n instead even if they are less interesting.
                if self.verbose is True:
                    print(
                        f"  g^i_n - g^i_m = {dg_0:.2f} kT, g^j_m - g^j_n = {dg_1:.2f} kT, Total dg: {dg:.2f} kT"
                    )

                prob_acc = min(1, np.exp(-dU + dg))
        return prob_acc

    def accept_or_reject(self, prob_acc):
        """
        Returns a boolean variable indiciating whether the proposed swap should be acceepted given the acceptance rate.

        Parameters
        ----------
        prob_acc : float
            The acceptance rate.

        Returns
        -------
        swap_bool : bool
            A boolean variable indicating whether the swap should be accepted.
        """
        if prob_acc == 0:
            swap_bool = False
            self.n_rejected += 1
            if self.verbose is True:
                print("  Swap rejected! ", end="", flush=True)
        else:
            rand = random.random()
            if self.verbose is True:
                print(
                    f"  Acceptance rate: {prob_acc:.3f} / Random number drawn: {rand:.3f}"
                )
            if rand < prob_acc:
                swap_bool = True
                if self.verbose is True:
                    print("  Swap accepted! ")
            else:
                swap_bool = False
                self.n_rejected += 1
                if self.verbose is True:
                    print("  Swap rejected! ")
        return swap_bool

    def histogram_correction(self, weights, counts):
        """
        Corrects the lambda weights based on the histogram counts. Namely,
        :math:`g_k' = g_k + ln(N_{k-1}/N_k)`, where :math:`g_k` and :math:`g_k'`
        are the lambda weight after and before the correction, respectively.
        Notably, in any of the following situations, we don't do any correction.

        - Either :math:`N_{k-1}` or :math:`N_k` is 0.
        - Either :math:`N_{k-1}` or :math:`N_k` is smaller than the histogram cutoff.

        Parameters
        ----------
        weights : list
            A list of lists of weights (of ALL simulations) to be corrected.
        counts : list
            A list of lists of counts (of ALL simulations).

        Return
        ------
        weights : list
            An updated list of lists of corected weights.
        """
        if self.N_cutoff == -1:
            print('\nNote: No histogram correction will be performed.')
        else:
            if self.verbose is True:
                print("\nPerforming histogram correction for the lambda weights ...")
            else:
                print("\nPerforming histogram correction for the lambda weights ...", end="")

            for i in range(len(weights)):  # loop over the replicas
                if self.verbose is True:
                    print(f"  Counts of rep {i}:\t\t{counts[i]}")
                    print(f'  Original weights of rep {i}:\t{[float(f"{k:.3f}") for k in weights[i]]}')

                for j in range(1, len(weights[i])):  # loop over the alchemical states
                    if counts[i][j - 1] != 0 and counts[i][j - 1] != 0:
                        if np.min([counts[i][j - 1], counts[i][j]]) > self.N_cutoff:
                            weights[i][j] += np.log(counts[i][j - 1] / counts[i][j])

                if self.verbose is True:
                    print(f'  Corrected weights of rep {i}:\t{[float(f"{k:.3f}") for k in weights[i]]}\n')

            if self.verbose is False:
                print(' Done')

        return weights

    def combine_weights(self, weights, method):
        """
        Combine alchemical weights across multiple replicas using probability ratios
        or weight differences. (See :ref:`doc_w_schemes` for mor details.)

        Parameters
        ----------
        weights : list
            A list of Wang-Landau weights of ALL simulations
        method : str
            Method for combining weights. Must be one of the following:

              * :code:`None`: No weight combination is performed, and the original weights are returned.
              * :code:`mean`: The arithmetic mean of the probability ratios is used to scale the weights.
              * :code:`geo-mean` or :code:`geo_mean`: The geometric mean of the probability ratios is
                used to scale the weights.
              * :code:`g-diff` or :code:`g_diff`: The difference between neighboring weights is used
                to determine the alchemical weights.

        Returns
        -------
        weights : list
            A list of original (if method is :code:`None`) or modified Wang-Landau weights of ALL simulations.
        g_vec : np.array
            An array of alchemical weights of the whole range of states.
        """
        if method is None:
            print('Note: No weight combination will be performed.')
            g_vec = None
            return weights, g_vec
        else:
            if self.verbose is True:
                print(f'Performing weight combination with the {method} method ...')
            else:
                print(f'Performing weight combination with the {method} method ...', end='')

            if self.verbose is True:
                w = np.round(weights, decimals=3).tolist()  # just for printing
                print('  Original weights:')
                for i in range(len(w)):
                    print(f'    Rep {i}: {w[i]}')

            if method == 'g-diff' or method == 'g_diff':
                # Method based on weight differences
                dg_vec = []
                dg_adjacent = [list(np.diff(weights[i])) for i in range(len(weights))]
                for i in range(self.n_tot - 1):
                    dg_list = []
                    for j in range(len(self.state_ranges)):
                        if i in self.state_ranges[j] and i + 1 in self.state_ranges[j]:
                            idx = self.state_ranges[j].index(i)
                            dg_list.append(dg_adjacent[j][idx])
                    dg_vec.append(np.mean(dg_list))
                dg_vec.insert(0, 0)
                g_vec = np.array([sum(dg_vec[:(i + 1)]) for i in range(len(dg_vec))])
            else:
                # Method based on probability ratios
                # Step 1: Convert the weights into probabilities
                weights = np.array(weights)
                prob = np.array([[np.exp(-i)/np.sum(np.exp(-weights[j])) for i in weights[j]] for j in range(len(weights))])  # noqa: E501

                # Step 2: Caclulate the probability ratios (after figuring out overlapped states between adjacent replicas)  # noqa: E501
                overlapped = [set(self.state_ranges[i]).intersection(set(self.state_ranges[i + 1])) for i in range(len(self.state_ranges) - 1)]  # noqa: E501
                prob_ratio = [prob[i + 1][: len(overlapped[i])] / prob[i][-len(overlapped[i]):] for i in range(len(overlapped))]  # noqa: E501

                # Step 3: Average the probability ratios
                avg_ratio = [1]   # This allows easier scaling since the first prob vector stays the same.
                if method == 'mean':
                    avg_ratio.extend([np.mean(prob_ratio[i]) for i in range(len(prob_ratio))])
                elif method == 'geo-mean':
                    avg_ratio.extend([np.prod(prob_ratio[i])**(1/len(prob_ratio[i])) for i in range(len(prob_ratio))])

                # Step 4: Scale the probabilities for each replica
                scaled_prob = np.array([prob[i] / np.prod(avg_ratio[: i + 1]) for i in range(len(prob))])

                # Step 5: Average and convert the probabilities
                p_vec = []
                for i in range(self.n_tot):   # global state index
                    p = []   # a list of probabilities to be averaged for each state
                    for j in range(len(self.state_ranges)):   # j can be regared as the replica index
                        if i in self.state_ranges[j]:
                            local_idx = i - j * self.s
                            p.append(scaled_prob[j][local_idx])
                    if method == 'mean':
                        p_vec.append(np.mean(p))
                    elif method == 'geo-mean' or method == 'geo_mean':
                        p_vec.append(np.prod(p) ** (1 / len(p)))

                g_vec = -np.log(p_vec)
                g_vec -= g_vec[0]

            # Determine the vector of alchemical weights for each replica
            weights = [list(g_vec[i: i + self.n_sub] - g_vec[i: i + self.n_sub][0]) for i in range(self.n_sim)]
            weights = np.round(weights, decimals=5).tolist()

            if self.verbose is True:
                w = np.round(weights, decimals=3).tolist()  # just for printing
                print('\n  Modified weights:')
                for i in range(len(w)):
                    print(f'    Rep {i}: {w[i]}')

            if self.verbose is False:
                print(' DONE')
                print(f'The alchemical weights of all states: \n  {list(np.round(g_vec, decimals=3))}')
            else:
                print(f'\n  The alchemical weights of all states: \n  {list(np.round(g_vec, decimals=3))}')

            return weights, g_vec

    def run_EEXE(self, n):
        """
        Makes TPR files and runs an ensemble of expanded ensemble simulations
        using GROMACS.

        Parameters
        ----------
        n : int
            The iteration index (starting from 0).

        Returns
        -------
        md : :code:`gmxapi.commandline.CommandlineOperation` obj
            The :code:`gmxapi.commandline.CommandlineOperation` object returned by :code:`gmxapi.mdrun`
            which contains STDOUT and STDERR of the simulation.

        Notes
        -----
        This function performs the following steps:

          1. Prepares TPR files for the simulation ensemble using :code:`grompp`.
          2. Runs all the simulations simultaneously using :code:`mdrun`.
          3. Removes any empty directories created by the function.

        The function assumes that the GROMACS executable is available.
        """
        if rank == 0:
            iter_str = f'\nIteration {n}: {self.dt * self.nst_sim * n: .1f} - {self.dt * self.nst_sim * (n + 1): .1f} ps'  # noqa: E501
            print(iter_str + '\n' + '=' * (len(iter_str) - 1))

        if rank == 0:
            dir_before = [
                i for i in os.listdir(".") if os.path.isdir(os.path.join(".", i))]
            print("Preparing the tpr files for the simulation ensemble ...", end="")

        grompp = gmx.commandline_operation(
            "gmx",
            arguments=["grompp"],  # noqa: E127
            input_files=[  # noqa: E127
                {
                    "-f": f"../sim_{i}/iteration_{n}/{self.mdp.split('/')[-1]}",
                    "-c": f"../sim_{i}/iteration_{n}/{self.gro.split('/')[-1]}",
                    "-p": f"../sim_{i}/iteration_{n}/{self.top.split('/')[-1]}",
                }
                for i in range(self.n_sim)
            ],
            output_files=[  # noqa: E127
                {  # noqa: E127
                    "-o": f"../sim_{i}/iteration_{n}/sys_EE.tpr",
                    "-po": f"../sim_{i}/iteration_{n}/mdout.mdp",
                }
                for i in range(self.n_sim)
            ],
        )
        grompp.run()
        if rank == 0:  # just print the messages once
            utils.gmx_output(grompp)

        # Run all the simulations simultaneously using gmxapi
        if rank == 0:
            print("Running an ensemble of simulations ...", end="")

        if self.parallel is True:
            tpr = [f'{grompp.output.file["-o"].result()[i]}' for i in range(self.n_sim)]
            inputs = gmx.read_tpr(tpr)
            md = gmx.mdrun(inputs, runtime_args=self.runtime_args)
            md.run()
        else:
            # Note that we could use output_files argument to customize the output file
            # names but here we'll just use the defaults.

            arguments = ['mdrun']  # arguments for gmx.commandline_operation

            if self.runtime_args is not None:
                # Turn the dictionary into a list with the keys alternating with values
                args_keys = list(self.runtime_args.keys())
                args_vals = list(self.runtime_args.values())

                add_args = []
                for i in range(len(args_keys)):
                    add_args.append(args_keys[i])
                    add_args.append(args_vals[i])

                arguments.extend(add_args)

            md = gmx.commandline_operation(
                "gmx",
                arguments=arguments,
                input_files=[  # noqa: E128
                    {
                        "-s": grompp.output.file["-o"].result()[i],
                    }
                    for i in range(self.n_sim)
                ],
            )
            md.run()
            if rank == 0:  # just print the messages once
                utils.gmx_output(md)

        if rank == 0:
            dir_after = [
                i for i in os.listdir(".") if os.path.isdir(os.path.join(".", i))
            ]
            utils.clean_up(dir_before, dir_after, self.verbose)

        return md



####################################################################
#                                                                  #
#    ensemble_md,                                                  #
#    a python package for running GROMACS simulation ensembles     #
#                                                                  #
#    Written by Wei-Tse Hsu <wehs7661@colorado.edu>                #
#    Copyright (c) 2022 University of Colorado Boulder             #
#                                                                  #
####################################################################
import os
import sys
import time
import glob
import pyemma
import pymbar
import pickle
import natsort
import argparse
import warnings
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import rc
from datetime import datetime
from deeptime.markov.tools.analysis import is_transition_matrix
warnings.simplefilter(action='ignore', category=UserWarning)

from ensemble_md.utils import utils  # noqa: E402
from ensemble_md.analysis import analyze_traj  # noqa: E402
from ensemble_md.analysis import analyze_matrix  # noqa: E402
from ensemble_md.analysis import msm_analysis  # noqa: E402
from ensemble_md.analysis import analyze_free_energy  # noqa: E402
from ensemble_md.ensemble_EXE import EnsembleEXE  # noqa: E402
from ensemble_md.utils.exceptions import ParameterError  # noqa: E402


def initialize(args):
    parser = argparse.ArgumentParser(
        description='This code analyzes an ensemble of expanded ensemble. Note that the template MDP\
                file specified in the YAML file needs to be available in the working directory.')
    parser.add_argument('-y',
                        '--yaml',
                        type=str,
                        default='params.yaml',
                        help='The input YAML file used to run the EEXE simulation. (Default: params.yaml)')
    parser.add_argument('-o',
                        '--output',
                        type=str,
                        default='analyze_EEXE_log.txt',
                        help='The output log file that contains the analysis results of EEXE. \
                            (Default: analyze_EEXE_log.txt)')
    parser.add_argument('-rt',
                        '--rep_trajs',
                        type=str,
                        default='rep_trajs.npy',
                        help='The NPY file containing the replica-space trajectory. (Default: rep_trajs.npy)')
    parser.add_argument('-st',
                        '--state_trajs',
                        type=str,
                        default='state_trajs.npy',
                        help='The NPY file containing the stitched state-space trajectory. \
                            If the specified file is not found, the code will try to find all the trajectories and \
                            stitch them. (Default: state_trajs.npy)')
    parser.add_argument('-d',
                        '--dir',
                        default='analysis',
                        help='The name of the folder for storing the analysis results.')
    parser.add_argument('-m',
                        '--maxwarn',
                        type=int,
                        default=0,
                        help='The maximum number of warnings in parameter specification to be ignored.')
    args_parse = parser.parse_args(args)

    return args_parse


def main():
    t0 = time.time()
    args = initialize(sys.argv[1:])
    warnings.simplefilter(action='ignore', category=RuntimeWarning)
    sys.stdout = utils.Logger(logfile=args.output)
    sys.stderr = utils.Logger(logfile=args.output)
    section_idx = 0

    rc('font', **{
        'family': 'sans-serif',
        'sans-serif': ['DejaVu Sans'],
    })
    # Set the font used for MathJax - more on this later
    rc('mathtext', **{'default': 'regular'})
    plt.rc('font', family='serif')

    print(f'Current time: {datetime.now().strftime("%d/%m/%Y %H:%M:%S")}')
    print(f'Command line: {" ".join(sys.argv)}')

    EEXE = EnsembleEXE(args.yaml)
    EEXE.print_params(params_analysis=True)

    for i in EEXE.warnings:
        print()
        print(f'{i}')
        print()

    if len(EEXE.warnings) > args.maxwarn:
        raise ParameterError(
            f"The execution failed due to warning(s) about parameter spcificaiton. Consider setting maxwarn in the input YAML file if you want to ignore them.")  # noqa: E501, F541

    # Check if the folder for saving the outputs exist. If not, create the folder.
    if os.path.exists(args.dir) is False:
        os.mkdir(args.dir)

    print('\nData analysis of the simulation ensemble')
    print('========================================')

    # Section 1. Analysis based on transitions between replicas
    print('[ Section 1. Analysis based on transitions between replicas ]')
    section_idx += 1

    # 1-0. Read in replica-space trajectories
    print('1-0. Reading in the replica-space trajectory ...')
    rep_trajs = np.load(args.rep_trajs)  # Shape: (n_sim, n_iter)

    # 1-1. Plot the replica-sapce trajectory
    print('1-1. Plotting transitions between replicas ...')
    dt_swap = EEXE.nst_sim * EEXE.dt    # dt for swapping replicas
    analyze_traj.plot_rep_trajs(rep_trajs, f'{args.dir}/rep_trajs.png', dt_swap)

    # 1-2. Plot the replica transition matrix
    print('1-2. Plotting the replica transition matrix (considering all configurations) ...')
    counts = [analyze_traj.traj2transmtx(rep_trajs[i], EEXE.n_sim, normalize=False) for i in range(len(rep_trajs))]
    reps_mtx = np.sum(counts, axis=0)  # First sum up the counts. This should be symmetric if n_ex=1. Otherwise it might not be. # noqa: E501
    reps_mtx /= np.sum(reps_mtx, axis=1)[:, None]   # and then normalize each row
    analyze_matrix.plot_matrix(reps_mtx, f'{args.dir}/rep_transmtx_allconfigs.png')

    # 1-3. Calculate the spectral gap for the replica transition amtrix
    spectral_gap = analyze_matrix.calc_spectral_gap(reps_mtx)
    print(f'1-3. The spectral gap of the replica transition matrix: {spectral_gap:.3f}')

    # Section 2. Analysis based on transitions between states
    print('\n[ Section 2. Analysis based on transitions between states ]')
    section_idx += 1

    # 2-0. Stitch the trajectories for each configuration
    if os.path.isfile(args.state_trajs) is True:
        print('2-0. Reading in the stitched state-space trajectory ...')
        state_trajs = np.load(args.state_trajs)
    else:
        # This may take a while.
        print('2-0. Stitching trajectories for each configuration from dhdl files ...')
        dhdl_files = [natsort.natsorted(glob.glob(f'sim_{i}/iteration_*/*dhdl*xvg')) for i in range(EEXE.n_sim)]
        shifts = np.arange(EEXE.n_sim) * EEXE.s
        state_trajs = analyze_traj.stitch_trajs(dhdl_files, rep_trajs, shifts=shifts)  # length: the number of replicas
        print('     Saving state_trajs.npy ...')
        np.save('state_trajs.npy', state_trajs)   # save the stithced trajectories

    # 2-1. Plot the state-space trajectory
    print('\n2-1. Plotting transitions between different alchemical states ...')
    dt_traj = EEXE.dt * EEXE.template['nstdhdl']  # in ps
    analyze_traj.plot_state_trajs(state_trajs, EEXE.state_ranges, f'{args.dir}/state_trajs.png', dt_traj)

    # 2-2. Plot the histograms for the states
    print('\n2-2. Plotting the histograms of the state index ...')
    analyze_traj.plot_state_hist(state_trajs, EEXE.state_ranges, f'{args.dir}/state_hist.png')

    # 2-3. Plot the overall state transition matrices calculated from the state-space trajectories
    print('\n2-3. Plotting the overall state transition matrices ...')
    mtx_list = []
    for i in range(EEXE.n_sim):
        mtx = analyze_traj.traj2transmtx(state_trajs[i], EEXE.n_tot)
        mtx_list.append(mtx)
        analyze_matrix.plot_matrix(mtx, f'{args.dir}/config_{i}_state_transmtx.png')

    # 2-4. For each configurration, calculate the spectral gap of the overall transition matrix obtained in step 2-2.
    print('\n2-4. Calculating the spectral gap of the state transition matrices ...')
    spectral_gaps = [analyze_matrix.calc_spectral_gap(mtx) for mtx in mtx_list]
    if None not in spectral_gaps:
        for i in range(EEXE.n_sim):
            print(f'   - Configuration {i}: {spectral_gaps[i]:.3f}')
        print(f'   - Average of the above: {np.mean(spectral_gaps):.3f} (std: {np.std(spectral_gaps, ddof=1):.3f})')

    # 2-5. For each configuration, calculate the stationary distribution from the overall transition matrix obtained in step 2-2.  # noqa: E501
    print('\n2-5. Calculating the stationary distributions ...')
    pi_list = [analyze_matrix.calc_equil_prob(mtx) for mtx in mtx_list]
    if any([x is None for x in pi_list]):
        pass  # None is in the list
    else:
        for i in range(EEXE.n_sim):
            print(f'   - Configuration {i}: {", ".join([f"{j:.3f}" for j in pi_list[i].reshape(-1)])}')
        if len({len(i) for i in pi_list}) == 1:  # all lists in pi_list have the same length
            print(f'   - Average of the above: {", ".join([f"{i:.3f}" for i in np.mean(pi_list, axis=0).reshape(-1)])}')  # noqa: E501

    # 2-6. Calculate the state index correlation time for each configuration (this step is more time-consuming one)
    print('\n2-6. Calculating the state index correlation time ...')
    tau_list = [(pymbar.timeseries.statistical_inefficiency(state_trajs[i], fast=True) - 1) / 2 * dt_traj for i in range(EEXE.n_sim)]  # noqa: E501
    for i in range(EEXE.n_sim):
        print(f'   - Configuration {i}: {tau_list[i]:.1f} ps')
    print(f'   - Average of the above: {np.mean(tau_list):.1f} ps (std: {np.std(tau_list, ddof=1):.1f} ps)')

    # 2-7. Calculate transit times for each configuration
    print('\n2-7. Plotting the average transit times ...')
    t_0k_list, t_k0_list, t_roundtrip_list, units = analyze_traj.plot_transit_time(state_trajs, EEXE.n_tot, dt=dt_traj, folder=args.dir)  # noqa: E501
    meta_list = [t_0k_list, t_k0_list, t_roundtrip_list]
    t_names = [
        '\n   - Average transit time from states 0 to k',
        '\n   - Average transit time from states k to 0',
        '\n   - Average round-trip time',
    ]
    for i in range(len(meta_list)):
        t_list = meta_list[i]
        print(t_names[i])
        for j in range(len(t_list)):
            if t_list[j] is None:
                print(f'     - Configuration {j}: Skipped')
            else:
                print(f'     - Configuration {j} ({len(t_list[j])} events): {np.mean(t_list[j]):.2f} {units}')
        print(f'     - Average of the above: {np.mean([np.mean(i) for i in t_list]):.2f} {units} (std: {np.std([np.mean(i) for i in t_list], ddof=1):.2f} {units})')  # noqa: E501

    if EEXE.msm is True:
        section_idx += 1

        # Section 3. Analysis based on Markov state models
        print('\n[ Section 3. Analysis based on Markov state models ]')

        # 3-1. Plot the implied timescale as a function of lag time
        print('\n3-1. Plotting the implied timescale as a function of lag time for all configurations ...')
        lags = np.arange(EEXE.lag_spacing, EEXE.lag_max + EEXE.lag_spacing, EEXE.lag_spacing)
        # lags could also be None and decided automatically. Could consider using that.
        ts_list = msm_analysis.plot_its(state_trajs, lags, fig_name=f'{args.dir}/implied_timescales.png', dt=dt_traj, units='ps')  # noqa: E501

        # 3-2. Decide a lag time for building MSM
        print('\n3-2. Calculating recommended lag times for building MSMs ...')
        chosen_lags = msm_analysis.decide_lagtimes(ts_list)
        print('     \nSuggested lag times for each configuration:')
        for i in range(len(chosen_lags)):
            print(f'       - Configuration {i}: {chosen_lags[i] * dt_traj:.1f} ps')

        # 3-3. Build a Bayesian MSM and perform a CK test for each configuration to validate the models
        print('\n3-3. Building Bayesian MSMs for the state-space trajectory for each configuration ...')
        print('     Performing a Chapman-Kolmogorov test on each trajectory ...')
        models = [pyemma.msm.bayesian_markov_model(state_trajs[i], chosen_lags[i], dt_traj=f'{dt_traj} ps', show_progress=False) for i in range(EEXE.n_sim)]  # noqa: E501

        for i in range(len(models)):
            print(f'     Plotting the CK-test results for configuration {i} ...')
            mlags = 5  # this maps to 5 (mlags: multiples of lag times for testing the model)
            nsets = models[i].nstates  # number of metastable states.
            # Note that nstates is the number of unique states in the input trajectores counted with the effective mode
            # (see the documentation) Therefore, if a system barely sampled some of the states, those states will
            # not be counted as involved in the transition matrix (i.e. not in the active set). To check the
            # active states, use models[i].active_set. If the system sampled all states frequently,
            # models[i].active_set should be equal to np.unique(state_trajs[i]) and both lengths should be
            # EEXE.n_tot. I'm not sure why the attribute nstates_full is not always EEXE.n_tot but is less
            # relevant here.
            cktest = models[i].cktest(nsets=nsets, mlags=mlags, show_progress=False)
            pyemma.plots.plot_cktest(cktest, dt=dt_traj, units='ps')
            plt.tight_layout(rect=[0, 0.03, 1, 0.98])
            plt.savefig(f'{args.dir}/config_{i}_CKtest.png', dpi=600)

        # Additionally, check if the sampling is poor for each configuration
        for i in range(len(models)):
            if models[i].nstates != EEXE.n_tot:
                print(f'     Note: The sampling of configuration {i} was poor.')

        # 3-4. Plot the state transition matrices estimated with the specified lag times in MSMs
        print('\n3-4. Plotting the overall transition matrices for all configurations and their average ...')
        mtx_list = [m.transition_matrix for m in models]
        mtx_list_modified = []  # just for plotting (if all trajs sampled the fulle range frequently, this will be the same as mtx_list)  # noqa: E501
        for i in range(len(mtx_list)):
            # check if each mtx in mtx_list spans the full alchemical range. (If the system did not visit
            # certain states, the dimension will be less than EEXE.n_tot * EEXE.n_tot. In this case, we
            # add rows and columns of 0. Note that the modified matrix will not be a transition matrix,
            # so this is only for plotting. For later analysis such as spectral gap calculation, we
            # will just use the unmodified matrices.
            if mtx_list[i].shape != (EEXE.n_tot, EEXE.n_tot):  # add rows and columns of 0
                sampled = models[i].active_set
                missing = list(set(range(EEXE.n_tot)) - set(sampled))  # states not visited

                # figure out which end we should stack rows/columns to
                n_1 = sum(missing > max(sampled))   # add rows/columns to the end of large state indices
                n_2 = sum(missing < min(sampled))   # add rows/columns to the end of small state indices

                # Below we first add a column and a row and so on
                mtx = mtx_list[i]
                for j in range(n_1):
                    d = mtx.shape[0]  # size of the square matrix
                    mtx = np.hstack((mtx, np.zeros([d, 1])))
                    mtx = np.vstack((mtx, np.zeros([1, d + 1])))
                for j in range(n_2):
                    d = mtx.shape[0]  # size of the square matrix
                    mtx = np.hstack((np.zeros([d, 1]), mtx))
                    mtx = np.vstack((np.zeros([1, d + 1]), mtx))
                mtx_list_modified.append(mtx)
            else:
                mtx_list_modified.append(mtx_list[i])

        avg_mtx = np.mean(mtx_list_modified, axis=0)
        print('     Saving transmtx.npy (plotted transition matrices)...')
        np.save(f'{args.dir}/transmtx.npy', mtx_list_modified)

        for i in range(EEXE.n_sim):
            analyze_matrix.plot_matrix(mtx_list[i], f'{args.dir}/config_{i}_state_transmtx_msm.png')
        analyze_matrix.plot_matrix(avg_mtx, f'{args.dir}/state_transmtx_avg_msm.png')

        # Note that if a configuration never visited a certain replica, the rows in the alchemical range of that
        # replica will only have 0 entries. Below we check if this is the case.
        print('     Checking the sum of each row of each transition matrix is 1 ...')
        is_transmtx = [is_transition_matrix(i) for i in mtx_list]
        for i in range(len(is_transmtx)):
            if is_transmtx[i] is False:
                print(f'     The sums of the rows are not 1 (but {np.sum(mtx_list[i], axis=1)}) for the transition matrix of configuration {i}.')  # noqa: E501

        # 3-5. Calculate the spectral gap from the transition matrix of each configuration
        print('\n3-5. Calculating the spectral gap of the state transition matrices obtained from MSMs ...')
        spectral_gaps = [analyze_matrix.calc_spectral_gap(mtx) for mtx in mtx_list]
        for i in range(EEXE.n_sim):
            print(f'       - Configuration {i}: {spectral_gaps[i]:.3f}')

        # 3-6. Calculate the stationary distribution for each configuration
        print('\n3-6. Calculating the stationary distributions from the transition matrices obtained from MSMs ...')
        pi_list = [m.pi for m in models]
        for i in range(EEXE.n_sim):
            print(f'       - Configuration {i}: {", ".join([f"{j:.3f}" for j in pi_list[i]])}')
        if len({len(i) for i in pi_list}) == 1:  # all lists in pi_list have the same length
            print(f'       - Average of the above: {", ".join([f"{i:.3f}" for i in np.mean(pi_list, axis=0)])}')

        # 3-7. Calculate the mean first-passage time (MFPT) from each MSM
        print('\n3-7. Calculating the mean first-passage times (MFPTs) ...')
        # note that it's not m.mfpt(min(m.active_set), max(m.active_set)) as the input to mfpt should be indices
        # though sometimes these two could be same.
        mfpt_list = [m.mfpt(0, m.nstates - 1) for m in models]
        for i in range(EEXE.n_sim):
            print(f'       - Configuration {i}: {mfpt_list[i]:.1f} ps')
        print(f'       - Average of the above: {np.mean(mfpt_list):.1f} ps (std: {np.std(mfpt_list, ddof=1):.1f} ps)')

        # 3-8. Calculate the state index correlation time for each configuration
        print('\n3-8. Plotting the state index correlation times for all configurations ...')
        msm_analysis.plot_acf(models, EEXE.n_tot, f'{args.dir}/state_ACF.png')

    # Section 4 (or Section 3). Free energy calculations
    if EEXE.free_energy is True:
        section_idx += 1
        print(f'\n[ Section {section_idx}. Free energy calculations ]')

        u_nk_list, dHdl_list = [], []

        if os.path.isfile(f'{args.dir}/u_nk_data.pickle') is True:
            print('Loading the preprocessed data u_nk ...')
            with open(f'{args.dir}/u_nk_data.pickle', 'rb') as handle:
                u_nk_list = pickle.load(handle)

        if os.path.isfile(f'{args.dir}/dHdl_data.pickle') is True:
            print('Loading the preprocessed data dHdl ...')
            with open(f'{args.dir}/dHdl_data.pickle', 'rb') as handle:
                dHdl_list = pickle.load(handle)

        if u_nk_list == [] and dHdl_list == []:
            for i in range(EEXE.n_sim):
                print(f'Reading dhdl files of replica {i} ...')
                files = natsort.natsorted(glob.glob(f'sim_{i}/iteration_*/*dhdl*xvg'))
                u_nk, dHdl = analyze_free_energy.preprocess_data(files, EEXE.temp, EEXE.df_spacing, EEXE.get_u_nk, EEXE.get_dHdl)  # noqa: E501
                u_nk_list.append(u_nk)
                dHdl_list.append(dHdl)

            with open(f'{args.dir}/u_nk_data.pickle', 'wb') as handle:
                pickle.dump(u_nk_list, handle, protocol=pickle.HIGHEST_PROTOCOL)

            with open(f'{args.dir}/dHdl_data.pickle', 'wb') as handle:
                pickle.dump(dHdl_list, handle, protocol=pickle.HIGHEST_PROTOCOL)

        state_ranges = [list(i) for i in EEXE.state_ranges]
        if EEXE.get_u_nk is True:
            f, f_err = analyze_free_energy.calculate_free_energy(u_nk_list, state_ranges, EEXE.df_method, EEXE.err_method, EEXE.n_bootstrap, EEXE.seed)  # noqa: E501
        else:
            f, f_err = analyze_free_energy.calculate_free_energy(dHdl_list, state_ranges, EEXE.df_method, EEXE.err_method, EEXE.n_bootstrap, EEXE.seed)  # noqa: E501

        print('Plotting the full-range free energy profile ...')
        analyze_free_energy.plot_free_energy(f, f_err, f'{args.dir}/free_energy_profile.png')

        print('The full-range free energy profile averaged over all replicas:')
        print(f"  {', '.join(f'{f[i]: .3f} +/- {f_err[i]: .3f} kT' for i in range(EEXE.n_tot))}")
        print(f'The free energy difference between the coupled and decoupled states: {f[-1]: .3f} +/- {f_err[-1]: .3f} kT')  # noqa: E501

    print(f'\nTime elpased: {utils.format_time(time.time() - t0)}')
